"""liquidctl drivers for sixth generation Asetek liquid coolers.

Copyright Andrew Robertson, Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import itertools
import logging

from liquidctl.driver.asetek import _Base690Lc
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

_READ_ENDPOINT = 0x81
_READ_MAX_LENGTH = 32
_WRITE_ENDPOINT = 0x1

_MAX_PROFILE_POINTS = 7

_CMD_READ_AIO_TEMP = 0xa9
_CMD_READ_FAN_SPEED = 0x41
_CMD_READ_FIRMWARE = 0xaa
_CMD_READ_PUMP_MODE = 0x33
_CMD_READ_PUMP_SPEED = 0x31
_CMD_WRITE_COLOR_LIST = 0x56
_CMD_WRITE_COLOR_SPEED = 0x53
_CMD_WRITE_FAN_CURVE = 0x40
_CMD_WRITE_FAN_SPEED = 0x42
_CMD_WRITE_PUMP_MODE = 0x32

_PUMP_MODES = ['quiet', 'balanced', 'performance']

_COLOR_SPEEDS = ['slower', 'normal', 'faster']

_COLOR_SPEEDS_VALUES = {
    'shift': [
        0x46, # slower
        0x28, # normal
        0x0F  # faster
    ],
    'pulse':[
        0x50, # slower
        0x37, # normal
        0x1E  # faster
    ],
    'blinking': [
        0x0F, # slower
        0x0A, # normal
        0x05  # faster
    ],
}

_COLOR_CHANGE_MODES = {
    'alert': [],
    'shift': [0x55, 0x01],
    'pulse': [0x52, 0x01],
    'blinking': [0x58, 0x01],
    'fixed': [0x55, 0x01],
}

# FIXME unknown required and maximum values
_COLOR_COUNT_BOUNDS = {
    'alert': (3, 3),
    'shift': (2, 4),
    'pulse': (1, 4),
    'blinking': (1, 4),
    'fixed': (1, 1),
}


def _quoted(*names):
    return ', '.join(map(repr, names))


# we inherit from _Base690Lc to reuse its implementation of connect
# and disconnect, that emulates the stock SiUSBXp driver on Windows
class HydroPro(_Base690Lc):
    """liquidctl driver for Corsair-branded sixth generation Asetek coolers."""

    _MATCHES = [
        (0x1b1c, 0x0c12, 'Corsair Hydro H150i Pro', {'fan_count': 3}),
        (0x1b1c, 0x0c13, 'Corsair Hydro H115i Pro', {'fan_count': 2}),
        (0x1b1c, 0x0c15, 'Corsair Hydro H100i Pro', {'fan_count': 2})
    ]

    def __init__(self, device, description, fan_count, **kwargs):
        super().__init__(device, description, **kwargs)
        self._fan_count = fan_count
        self._data = None

    def _post(self, data, *, read_length=None):
        """Write `data` and return response of up to `read_length` bytes."""

        assert read_length is not None and read_length <= _READ_MAX_LENGTH

        self.device.write(_WRITE_ENDPOINT, data)
        return self.device.read(_READ_ENDPOINT, read_length)[0:read_length]

    def initialize(self, pump_mode='balanced', **kwargs):
        """Initialize the device."""

        pump_mode = pump_mode.lower()

        if pump_mode not in _PUMP_MODES:
            raise ValueError(f'unknown pump mode, should be one of: {_quoted(*_PUMP_MODES)}')

        self._post([_CMD_WRITE_PUMP_MODE, _PUMP_MODES.index(pump_mode)], read_length=5)

        msg = self._post([_CMD_READ_FIRMWARE], read_length=7)
        firmware = '{}.{}.{}.{}'.format(*tuple(msg[3:7]))

        self.device.release()

        return [
            ('Firmware version', firmware, ''),
        ]

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        msg = self._post([_CMD_READ_AIO_TEMP], read_length=6)
        aio_temp = msg[3] + msg[4]/10

        fan_speeds = self._get_fan_speeds()

        msg = self._post([_CMD_READ_PUMP_MODE], read_length=4)
        pump_mode = _PUMP_MODES[msg[3]]

        msg = self._post([_CMD_READ_PUMP_SPEED], read_length=5)
        pump_speed = (msg[3] << 8) + msg[4]

        self.device.release()

        status = [('Liquid temperature', aio_temp, '°C')]

        for fan, speed in fan_speeds:
            status.append((f'Fan {fan} speed', speed, 'rpm'))

        return status + [
            ('Pump mode', pump_mode, ""),
            ('Pump speed', pump_speed, 'rpm'),
        ]

    def _get_fan_speeds(self):
        """Read the RPM speed of the fans."""

        speeds = []

        for i in range(self._fan_count):
            msg = self._post([_CMD_READ_FAN_SPEED, i], read_length=6)

            if msg[0] != 0x41 or msg[3] != i:
                _LOGGER.warning('failed to get current speed of fan %d', i)
                continue

            speeds.append((i + 1, (msg[4] << 8) + msg[5]))

        return speeds

    def set_color(self, channel, mode, colors, speed='normal', **kwargs):
        """Set the color mode for a specific channel."""

        mode = mode.lower()
        speed = speed.lower()
        colors = list(colors)

        if mode not in _COLOR_CHANGE_MODES:
            valid = _quoted(*_COLOR_CHANGE_MODES.keys())
            raise ValueError(f'unknown lighting mode, should be one of: {valid}')

        if speed not in _COLOR_SPEEDS:
            valid = _quoted(*_COLOR_SPEEDS)
            raise ValueError(f'unknown speed value, should be one of {valid}')

        if mode == 'alert':
            # FIXME this mode is far from being completely implemented; for
            # one, the temperatures are hardcoded; additionally, it may also be
            # possible to combine it with other modes, but exploring that would
            # require some experimentation
            temps = (30, 40, 50)
            self._post([0x5f, temps[0], 0x00, temps[1], 0x00, temps[2], 0x00]
                       + colors[0] + colors[1] + colors[2], read_length=6)
            self._post([0x5e, 0x01], read_length=3)
            self.device.release()
            return

        colors = self._check_color_count_bounds(colors, mode)

        if mode == 'fixed':
            colors = [colors[0], colors[0]]

        set_colors = list(itertools.chain(*colors))
        self._post([_CMD_WRITE_COLOR_LIST, len(colors)] + set_colors, read_length=3)

        if mode != 'fixed':
            magic_value = _COLOR_SPEEDS_VALUES[mode][_COLOR_SPEEDS.index(speed)]
            self._post([_CMD_WRITE_COLOR_SPEED, magic_value], read_length=3)

        self._post(_COLOR_CHANGE_MODES[mode], read_length=3)
        self.device.release()

    def _check_color_count_bounds(self, color_list, mode_name):
        requires, maximum = _COLOR_COUNT_BOUNDS[mode_name]

        if len(color_list) < requires:
            raise ValueError(f'{mode_name} mode requires {requires} colors')

        if len(color_list) > maximum:
            _LOGGER.debug('too many colors, dropping to %d', maximum)
            color_list = color_list[:maximum]

        return color_list

    def set_speed_profile(self, channel, profile, **kwargs):
        """Set channel to follow a speed duty profile."""

        channel = channel.lower()
        fan_indexes = self._fan_indexes(channel)

        adjusted = self._prepare_profile(profile, 0, 100, _MAX_PROFILE_POINTS)
        for temp, duty in adjusted:
            _LOGGER.info('setting %s PWM point: (%i°C, %i%%), device interpolated',
                        channel, temp, duty)

        temps, duties = map(list, zip(*adjusted))
        for i in fan_indexes:
            self._post([_CMD_WRITE_FAN_CURVE, i] + temps + duties, read_length=32)

        self.device.release()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""

        channel = channel.lower()
        duty = clamp(duty, 0, 100)

        for i in self._fan_indexes(channel):
            _LOGGER.info('setting speed for fan %d to %d', i + 1, duty)
            self._post([_CMD_WRITE_FAN_SPEED, i, duty], read_length=32)
        self.device.release()

    def _fan_indexes(self, channel):
        if channel.startswith('fan'):
            if len(channel) > 3:
                channel_num = int(channel[3:]) - 1
                if channel_num >= self._fan_count:
                    raise ValueError(f'unknown channel: {channel}')
                return [channel_num]
            return range(self._fan_count)
        elif channel == 'pump':
            raise NotSupportedByDevice()
        else:
            raise ValueError(f'unknown channel: {channel}')

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    @classmethod
    def probe(cls, handle, **kwargs):
        return super().probe(handle, **kwargs)


# backward compatibility
CorsairAsetekProDriver = HydroPro



# uses the psf/black style

import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.smart_device import SmartDevice2

# https://github.com/liquidctl/liquidctl/issues/292#issuecomment-786876335
# (adapted: set control mode for connected fan to PWM)
SAMPLE_STATUS = bytes.fromhex(
    "67023a003f00185732533230312003000200000000000000fc03000000000000"
    "0000000000000000322828000000000032282800000000003000000000000000"
)


class MockSmart2(MockHidapiDevice):
    def __init__(self, raw_speed_channels, raw_led_channels):
        super().__init__()
        self.raw_speed_channels = raw_speed_channels
        self.raw_led_channels = raw_led_channels

    def write(self, data):
        reply = bytearray(64)
        if data[0:2] == [0x10, 0x01]:
            reply[0:2] = [0x11, 0x01]
        elif data[0:2] == [0x20, 0x03]:
            reply[0:2] = [0x21, 0x03]
            reply[14] = self.raw_led_channels
            if self.raw_led_channels > 1:
                reply[15 + 1 * 6] = 0x10
                reply[15 + 2 * 6] = 0x11
        self.preload_read(Report(reply[0], reply[1:]))
        return super().write(data)


@pytest.fixture
def mock_smart2():
    raw = MockSmart2(raw_speed_channels=3, raw_led_channels=2)
    dev = SmartDevice2(raw, "Mock Smart Device V2", speed_channel_count=3, color_channel_count=2)
    dev.connect()
    return dev


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True), (True, False)])
def test_initializes(mock_smart2, has_hwmon, direct_access, tmp_path):
    if has_hwmon:
        mock_smart2._hwmon = HwmonDevice("mock_module", tmp_path)

    # TODO check the result
    _ = mock_smart2.initialize(direct_access=direct_access)

    writes = len(mock_smart2.device.sent)
    if not has_hwmon or direct_access:
        assert writes == 4
    else:
        assert writes == 2


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_reads_status_directly(mock_smart2, has_hwmon, direct_access):
    if has_hwmon:
        mock_smart2._hwmon = HwmonDevice(None, None)

    mock_smart2.device.preload_read(Report(0, SAMPLE_STATUS))

    expected = [
        ("Fan 1 speed", 1020, "rpm"),
        ("Fan 1 duty", 50, "%"),
        ("Fan 1 control mode", "PWM", ""),
        ("Fan 2 speed", 0, "rpm"),
        ("Fan 2 duty", 40, "%"),
        ("Fan 2 control mode", None, ""),
        ("Fan 3 speed", 0, "rpm"),
        ("Fan 3 duty", 40, "%"),
        ("Fan 3 control mode", None, ""),
        ("Noise level", 48, "dB"),
    ]

    got = mock_smart2.get_status(direct_access=direct_access)

    assert sorted(got) == sorted(expected)


def test_reads_status_from_hwmon(mock_smart2, tmp_path):
    mock_smart2._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "pwm1_enable").write_text("1\n")
    (tmp_path / "pwm2_enable").write_text("0\n")
    (tmp_path / "pwm3_enable").write_text("0\n")
    (tmp_path / "pwm1_mode").write_text("1\n")
    (tmp_path / "pwm2_mode").write_text("0\n")
    (tmp_path / "pwm3_mode").write_text("0\n")
    (tmp_path / "pwm1").write_text("127\n")
    (tmp_path / "pwm2").write_text("102\n")
    (tmp_path / "pwm3").write_text("102\n")
    (tmp_path / "fan1_input").write_text("1020\n")
    (tmp_path / "fan2_input").write_text("0\n")
    (tmp_path / "fan3_input").write_text("0\n")

    expected = [
        ("Fan 1 speed", 1020, "rpm"),
        ("Fan 1 duty", pytest.approx(50, rel=1.0 / 255), "%"),
        ("Fan 1 control mode", "PWM", ""),
        ("Fan 2 speed", 0, "rpm"),
        ("Fan 2 duty", pytest.approx(40, rel=1.0 / 255), "%"),
        ("Fan 2 control mode", "DC", ""),
        ("Fan 3 speed", 0, "rpm"),
        ("Fan 3 duty", pytest.approx(40, rel=1.0 / 255), "%"),
        ("Fan 3 control mode", "DC", ""),
    ]

    got = mock_smart2.get_status()

    assert sorted(got) == sorted(expected)


def test_constructor_sets_up_all_channels(mock_smart2):
    assert mock_smart2._speed_channels == {
        "fan1": (0, 0, 100),
        "fan2": (1, 0, 100),
        "fan3": (2, 0, 100),
    }
    assert mock_smart2._color_channels == {
        "led1": (0b001),
        "led2": (0b010),
        "sync": (0b011),
    }


def test_not_totally_broken(mock_smart2):
    _ = mock_smart2.initialize()
    mock_smart2.device.preload_read(Report(0, [0x67, 0x02] + [0] * 62))
    _ = mock_smart2.get_status()
    mock_smart2.set_color(
        channel="led1", mode="breathing", colors=iter([[142, 24, 68]]), speed="fastest"
    )
    mock_smart2.set_fixed_speed(channel="fan3", duty=50)



"""liquidctl driver for Gigabyte RGB Fusion 2.0 USB controllers.

Supported controllers:

- ITE 5702: found in Gigabyte Z490 Vision D
- ITE 8297: found in Gigabyte X570 Aorus Elite

Copyright CaseySJ, Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import sys
from collections import namedtuple

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

_USAGE_PAGE = 0xff89
_RGB_CONTROL_USAGE = 0xcc
_OTHER_USAGE = 0x10
_REPORT_ID = 0xcc
_REPORT_BYTE_LENGTH = 63
_INIT_CMD = 0x60

_COLOR_CHANNELS = {
    'led1': (0x20, 0x01),
    'led2': (0x21, 0x02),
    'led3': (0x22, 0x04),
    'led4': (0x23, 0x08),
    'led5': (0x24, 0x10),
    'led6': (0x25, 0x20),
    'led7': (0x26, 0x40),
    'led8': (0x27, 0x80),
}

_PULSE_SPEEDS = {
    'slowest':                          (0x40, 0x06, 0x40, 0x06, 0x20, 0x03),
    'slower':                           (0x78, 0x05, 0x78, 0x05, 0xbc, 0x02),
    'normal':                           (0xb0, 0x04, 0xb0, 0x04, 0xf4, 0x01),
    'faster':                           (0xe8, 0x03, 0xe8, 0x03, 0xf4, 0x01),
    'fastest':                          (0x84, 0x03, 0x84, 0x03, 0xc2, 0x01),
    'ludicrous':                        (0x20, 0x03, 0x20, 0x03, 0x90, 0x01),
}

_FLASH_SPEEDS = {
    'slowest':                          (0x64, 0x00, 0x64, 0x00, 0x60, 0x09),
    'slower':                           (0x64, 0x00, 0x64, 0x00, 0x90, 0x08),
    'normal':                           (0x64, 0x00, 0x64, 0x00, 0xd0, 0x07),
    'faster':                           (0x64, 0x00, 0x64, 0x00, 0x08, 0x07),
    'fastest':                          (0x64, 0x00, 0x64, 0x00, 0x40, 0x06),
    'ludicrous':                        (0x64, 0x00, 0x64, 0x00, 0x78, 0x05),
}

_DOUBLE_FLASH_SPEEDS = {
    'slowest':                          (0x64, 0x00, 0x64, 0x00, 0x28, 0x0a),
    'slower ':                          (0x64, 0x00, 0x64, 0x00, 0x60, 0x09),
    'normal':                           (0x64, 0x00, 0x64, 0x00, 0x90, 0x08),
    'faster':                           (0x64, 0x00, 0x64, 0x00, 0xd0, 0x07),
    'fastest':                          (0x64, 0x00, 0x64, 0x00, 0x08, 0x07),
    'ludicrous':                        (0x64, 0x00, 0x64, 0x00, 0x40, 0x06),
}

_COLOR_CYCLE_SPEEDS = {
    'slowest':                          (0x78, 0x05, 0xb0, 0x04, 0x00, 0x00),
    'slower':                           (0x7e, 0x04, 0x1a, 0x04, 0x00, 0x00),
    'normal':                           (0x52, 0x03, 0xee, 0x02, 0x00, 0x00),
    'faster':                           (0xf8, 0x02, 0x94, 0x02, 0x00, 0x00),
    'fastest':                          (0x26, 0x02, 0xc2, 0x01, 0x00, 0x00),
    'ludicrous':                        (0xcc, 0x01, 0x68, 0x01, 0x00, 0x00),
}

_ColorMode = namedtuple('_ColorMode', ['name', 'value', 'pulses', 'flash_count',
                                       'cycle_count', 'max_brightness', 'takes_color',
                                       'speed_values'])

_COLOR_MODES = {
    mode.name: mode
    for mode in [
        _ColorMode('off', 0x01, pulses=False, flash_count=0, cycle_count=0,
                   max_brightness=0, takes_color=False, speed_values=None),
        _ColorMode('fixed', 0x01, pulses=False, flash_count=0, cycle_count=0,
                   max_brightness=90, takes_color=True, speed_values=None),
        _ColorMode('pulse', 0x02, pulses=True, flash_count=0, cycle_count=0,
                   max_brightness=90, takes_color=True, speed_values=_PULSE_SPEEDS),
        _ColorMode('flash', 0x03, pulses=True, flash_count=1, cycle_count=0,
                   max_brightness=100, takes_color=True, speed_values=_FLASH_SPEEDS),
        _ColorMode('double-flash', 0x03, pulses=True, flash_count=2, cycle_count=0,
                   max_brightness=100, takes_color=True, speed_values=_DOUBLE_FLASH_SPEEDS),
        _ColorMode('color-cycle', 0x04, pulses=False, flash_count=0, cycle_count=7,
                   max_brightness=100, takes_color=False, speed_values=_COLOR_CYCLE_SPEEDS),
    ]
}


class RgbFusion2(UsbHidDriver):
    """liquidctl driver for Gigabyte RGB Fusion 2.0 USB controllers."""

    _MATCHES = [
        (0x048d, 0x5702, 'Gigabyte RGB Fusion 2.0 5702 Controller', {}),
        (0x048d, 0x8297, 'Gigabyte RGB Fusion 2.0 8297 Controller', {}),
    ]

    @classmethod
    def probe(cls, handle, **kwargs):
        """Probe `handle` and yield corresponding driver instances.

        These devices have multiple top-level HID usages, and HidapiDevice
        handles matching other usages have to be ignored.
        """

        # if usage_page/usage are not available due to hidapi limitations
        # (version, platform or backend), they are unfortunately left
        # uninitialized; because of this, we explicitly exclude the undesired
        # usage_page/usage pair, and assume in all other cases that we either
        # have the desired usage page/usage pair, or that on that system a
        # single handle is returned for that device interface (see: 259)

        if (handle.hidinfo['usage_page'] == _USAGE_PAGE and
                handle.hidinfo['usage'] == _OTHER_USAGE):
            return

        yield from super().probe(handle, **kwargs)

    def initialize(self, **kwargs):
        """Initialize the device.

        Returns a list of `(property, value, unit)` tuples, containing the
        firmware version and other useful information provided by the hardware.
        """

        self._send_feature_report([_REPORT_ID, _INIT_CMD])
        data = self._get_feature_report(_REPORT_ID)
        # be tolerant: 8297 controllers support report IDs yet return 0 in the
        # first byte, which is out of spec
        assert data[0] in (_REPORT_ID, 0) and data[1] == 0x01

        null = data.index(0, 12)
        dev_name = str(bytes(data[12:null]), 'ascii', errors='replace')
        fw_version = tuple(data[4:8])
        return [
            ('Hardware name', dev_name, ''),
            ('Firmware version', '{}.{}.{}.{}'.format(*fw_version), ''),
        ]

    def get_status(self, **kwargs):
        """Get a status report.

        Currently returns an empty list, but this behavior is not guaranteed as
        in the future the device may start to report useful information.  A
        non-empty list would contain `(property, value, unit)` tuples.
        """

        _LOGGER.info('status reports not available from %s', self.description)
        return []

    def set_color(self, channel, mode, colors, speed='normal', **kwargs):
        """Set the color mode for a specific channel.

        Up to eight individual channels are available, named 'led1' through
        'led8'.  In addition to these, the 'sync' channel can be used to apply
        the same settings to all channels.

        The table bellow summarizes the available channels.

        | Mode         | Colors required | Speed is customizable |
        | ------------ | --------------- | --------------------- |
        | off          |            zero |                    no |
        | fixed        |             one |                    no |
        | pulse        |             one |                   yes |
        | flash        |             one |                   yes |
        | double-flash |             one |                   yes |
        | color-cycle  |            zero |                   yes |

        `colors` should be an iterable of zero or one `[red, blue, green]`
        triples, where each red/blue/green component is a value in the range
        0–255.

        `speed`, when supported by the `mode`, can be one of: `slowest`,
        `slower`, `normal` (default), `faster`, `fastest` or `ludicrous`.
        """

        mode = _COLOR_MODES[mode]
        colors = iter(colors)

        if mode.takes_color:
            try:
                r, g, b = next(colors)
                single_color = (b, g, r)
            except StopIteration:
                raise ValueError(f'one color required for mode={mode.name}') from None
        else:
            single_color = (0, 0, 0)
        remaining = sum(1 for _ in colors)
        if remaining:
            _LOGGER.warning('too many colors for mode=%s, dropping %d', mode.name, remaining)

        brightness = clamp(100, 0, mode.max_brightness)  # hardcode this for now
        data = [_REPORT_ID, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, mode.value, brightness, 0x00]
        data += single_color
        data += [0x00, 0x00, 0x00, 0x00, 0x00]
        if mode.speed_values:
            data += mode.speed_values[speed]
        else:
            data += [0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        data += [0x00, 0x00, mode.cycle_count, int(mode.pulses), mode.flash_count]

        if channel == 'sync':
            selected_channels = _COLOR_CHANNELS.values()
        else:
            selected_channels = (_COLOR_CHANNELS[channel],)
        for addr1, addr2 in selected_channels:
            data[1:3] = addr1, addr2
            self._send_feature_report(data)
        self._execute_report()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def reset_all_channels(self):
        """Reset all LED channels."""
        for addr1, _ in _COLOR_CHANNELS.values():
            self._send_feature_report([_REPORT_ID, addr1, 0])
        self._execute_report()

    def _get_feature_report(self, report_id):
        return self.device.get_feature_report(report_id, _REPORT_BYTE_LENGTH + 1)

    def _send_feature_report(self, data):
        padding = [0x0]*(_REPORT_BYTE_LENGTH + 1 - len(data))
        self.device.send_feature_report(data + padding)

    def _execute_report(self):
        """Request for the previously sent lighting settings to be applied."""
        self._send_feature_report([_REPORT_ID, 0x28, 0xff])

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


# Acknowledgments:
#
# Thanks to SgtSixPack for capturing USB traffic on 0x8297 and testing the driver on Windows.



"""Test API backward compatibility with liquidctl 1.0.x.

While at the time all APIs were undocumented, we choose to support the use
cases from GKraken as that software is a substantial contribution to the
community.
"""

import pytest
from test_kraken2 import mockKrakenXDevice

from liquidctl.version import __version__

SPECTRUM = [
    (235, 77, 40),
    (255, 148, 117),
    (126, 66, 45),
    (165, 87, 0),
    (56, 193, 66),
    (116, 217, 170),
    (166, 158, 255),
    (208, 0, 122)
]


def test_pre11_apis_deprecated_super_mode(mockKrakenXDevice):
    # deprecated in favor of super-fixed, super-breathing and super-wave
    mockKrakenXDevice.set_color('sync', 'super', [(128, 0, 255)] + SPECTRUM, 'normal')



"""Constants and methods for interfacing with PMBus compliant devices.

Specifications:

- Power Systems Management Protocol Specification.  Revision 1.3.1, 2015.
  Available uppon request, check the PMBus website.

- Power Systems Management Protocol Specification.  Revision 1.2, 2010.
  Available on the PMBus website.
  http://pmbus.org/Assets/PDFS/Public/PMBus_Specification_Part_I_Rev_1-2_20100906.pdf
  http://pmbus.org/Assets/PDFS/Public/PMBus_Specification_Part_II_Rev_1-2_20100906.pdf

- System Management Bus (SMBus) Specification.  Version 3.1, 2018.
  Available on the SMBus website.
  http://smbus.org/specs/SMBus_3_1_20180319.pdf

Additional references:

- Milios, John.  CRC-8 firmware implementations for SMBus.  1999.
 http://sbs-forum.org/marcom/dc2/20_crc-8_firmware_implementations.pdf

- Pircher, Thomas.  pycrc -- parameterisable CRC calculation utility and C
  source code generator: CRC algorithms implemented in Python.
  https://github.com/tpircher/pycrc/blob/master/pycrc/algorithms.py

- White, Robert V.  Using the PMBus Protocol.  2005.
  http://pmbus.org/Assets/Present/Using_The_PMBus_20051012.pdf

Copyright Jonas Malaco and contributors

Includes a CRC-8 implementation adapted from pycrc by Thomas Pircher.
Copyright (c) 2006-2017  Thomas Pircher  <tehpeh-web@tty1.net>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import math
from enum import IntEnum, IntFlag, unique

from liquidctl.util import mkCrcFun


@unique
class WriteBit(IntFlag):
    WRITE = 0x00
    READ = 0x01


@unique
class CommandCode(IntEnum):
    """Incomplete enumeration of the PMBus command codes."""

    PAGE = 0x00

    CLEAR_FAULTS = 0x03

    PAGE_PLUS_WRITE = 0x05
    PAGE_PLUS_READ = 0x06

    VOUT_MODE = 0x20

    FAN_CONFIG_1_2 = 0x3a
    FAN_COMMAND_1 = 0x3b
    FAN_COMMAND_2 = 0x3c
    FAN_CONFIG_3_4 = 0x3d
    FAN_COMMAND_3 = 0x3e
    FAN_COMMAND_4 = 0x3f

    READ_EIN = 0x86
    READ_EOUT = 0x87
    READ_VIN = 0x88
    READ_IIN = 0x89
    READ_VCAP = 0x8a
    READ_VOUT = 0x8b
    READ_IOUT = 0x8c
    READ_TEMPERATURE_1 = 0x8d
    READ_TEMPERATURE_2 = 0x8e
    READ_TEMPERATURE_3 = 0x8f
    READ_FAN_SPEED_1 = 0x90
    READ_FAN_SPEED_2 = 0x91
    READ_FAN_SPEED_3 = 0x92
    READ_FAN_SPEED_4 = 0x93
    READ_DUTY_CYCLE = 0x94
    READ_FREQUENCY = 0x95
    READ_POUT = 0x96
    READ_PIN = 0x97
    READ_PMBUS_REVISON = 0x98
    MFR_ID = 0x99
    MFR_MODEL = 0x9a
    MFR_REVISION = 0x9b
    MFR_LOCATION = 0x9c
    MFR_DATE = 0x9d
    MFR_SERIAL = 0x9e

    MFR_SPECIFIC_D1 = 0xd1
    MFR_SPECIFIC_D2 = 0xd2
    MFR_SPECIFIC_D8 = 0xd8
    MFR_SPECIFIC_DC = 0xdc
    MFR_SPECIFIC_EE = 0xee
    MFR_SPECIFIC_F0 = 0xf0
    MFR_SPECIFIC_FC = 0xfc


def linear_to_float(bytes, vout_exp=None):
    """Read PMBus LINEAR11 and ULINEAR16 numeric values.

    If `vout_exp` is None the value is interpreted as a 2 byte LINEAR11 value.
    The mantissa is stored in the lower 11 bits, in two's-complement, and the
    exponent is is stored in the upper 5 bits, also in two's-complement.

    Otherwise the value is assumed to be encoded in ULINEAR16, where the
    exponent is read from the lower 5 bits of `vout_exp` (which is assumed to
    be the output from VOUT_MOE) and the mantissa is the unsigned 2 byte
    integer in `bytes`.

    Per the SMBus specification, the lowest order byte is sent first (endianess
    is little).

    >>> linear_to_float(bytes.fromhex('67e3'))
    54.4375
    >>> linear_to_float(bytes.fromhex('6703'), vout_exp=0x1c)
    54.4375
    """
    tmp = int.from_bytes(bytes[:2], byteorder='little')
    if vout_exp is None:
        exp = tmp >> 11
        fra = tmp & 0x7ff
        if fra > 1023:
            fra = fra - 2048
    else:
        exp = vout_exp & 0x1f
        fra = tmp
    if exp > 15:
        exp = exp - 32
    return fra * 2**exp


def float_to_linear11(float):
    """Encode float in PMBus LINEAR11 format.

    A LINEAR11 number is a 2 byte value with an 11 bit two's complement
    mantissa and a 5 bit two's complement exponent.

    Per the SMBus specification, the lowest order byte is sent first (endianess
    is little).

    >>> float_to_linear11(3.3).hex()
    '4dc3'
    >>> float_to_linear11(0.0).hex()
    '0000'
    >>> linear_to_float(float_to_linear11(2812))
    2812
    >>> linear_to_float(float_to_linear11(-2812))
    -2812
    """
    if float == 0:
        return b'\x00\x00'
    max_y = 1023
    n = math.ceil(math.log(math.fabs(float)/max_y, 2))
    y = round(float * 2**(-n))
    if n < 0:
        n = n + 32
    if y < 0:
        y = y + 2048
    return int.to_bytes((n << 11) | y, length=2, byteorder='little')


def compute_pec(bytes):
    """
    Compute a 8-bit Packet Error Code (PEC) for `bytes`.

    According to the SMBus specification, the PEC is computed using a 8-bit
    cyclic rendundancy check (CRC-8) with the polynominal x⁸ + x² + x¹ + x⁰.

    The computation uses a 256-byte lookup table.

    Based on https://github.com/tpircher/pycrc/blob/master/pycrc/algorithms.py.

    >>> hex(compute_pec(bytes('123456789', 'ascii')))
    '0xf4'
    >>> hex(compute_pec(bytes.fromhex('5c')))
    '0x93'
    >>> hex(compute_pec(bytes.fromhex('5c93')))
    '0x0'
    """

    return mkCrcFun('crc-8')(bytes)



import pytest
from _testutils import MockHidapiDevice

from liquidctl.driver.usb import UsbDriver, UsbHidDriver


@pytest.fixture
def emulated_hid_device():
    hiddev = MockHidapiDevice()
    return UsbHidDriver(hiddev, 'Test')


@pytest.fixture
def emulated_usb_device():
    usbdev = MockHidapiDevice()  # hack, should mock PyUsbDevice
    dev = UsbDriver(usbdev, 'Test')

    return dev


def test_hid_connects(emulated_hid_device):
    dev = emulated_hid_device

    def mock_open():
        nonlocal opened
        opened = True

    dev.device.open = mock_open
    opened = False

    with dev.connect() as cm:
        assert cm == dev
        assert opened


def test_hid_disconnect(emulated_hid_device):
    dev = emulated_hid_device

    def mock_close():
        nonlocal opened
        opened = False

    dev.device.close = mock_close
    opened = True

    dev.disconnect()
    assert not opened


def test_usb_connects(emulated_usb_device):
    dev = emulated_usb_device

    def mock_open():
        nonlocal opened
        opened = True

    dev.device.open = mock_open
    opened = False

    with dev.connect() as cm:
        assert cm == dev
        assert opened


def test_usb_disconnect(emulated_usb_device):
    dev = emulated_usb_device

    def mock_close():
        nonlocal opened
        opened = False

    dev.device.close = mock_close
    opened = True

    dev.disconnect()
    assert not opened



import pytest
from _testutils import MockHidapiDevice

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.kraken2 import Kraken2
from liquidctl.error import NotSupportedByDevice


@pytest.fixture
def mockKrakenXDevice():
    device = _MockKrakenDevice(fw_version=(6, 0, 2))
    dev = Kraken2(device, 'Mock X62', device_type=Kraken2.DEVICE_KRAKENX)

    dev.connect()
    return dev


@pytest.fixture
def mockOldKrakenXDevice():
    device = _MockKrakenDevice(fw_version=(2, 1, 8))
    dev = Kraken2(device, 'Mock X62', device_type=Kraken2.DEVICE_KRAKENX)

    dev.connect()
    return dev


@pytest.fixture
def mockKrakenMDevice():
    device = _MockKrakenDevice(fw_version=(6, 0, 2))
    dev = Kraken2(device, 'Mock M22', device_type=Kraken2.DEVICE_KRAKENM)

    dev.connect()
    return dev


class _MockKrakenDevice(MockHidapiDevice):
    def __init__(self, fw_version):
        super().__init__(vendor_id=0xffff, product_id=0x1e71)
        self.fw_version = fw_version
        self.temperature = 30.9
        self.fan_speed = 1499
        self.pump_speed = 2702

    def read(self, length):
        pre = super().read(length)
        if pre:
            return pre
        buf = bytearray(64)
        buf[1:3] = divmod(int(self.temperature * 10), 10)
        buf[3:5] = self.fan_speed.to_bytes(length=2, byteorder='big')
        buf[5:7] = self.pump_speed.to_bytes(length=2, byteorder='big')
        major, minor, patch = self.fw_version
        buf[0xb] = major
        buf[0xc:0xe] = minor.to_bytes(length=2, byteorder='big')
        buf[0xe] = patch
        return buf[:length]


def test_kraken_connect(mockKrakenXDevice):
    def mock_open():
        nonlocal opened
        opened = True

    mockKrakenXDevice.device.open = mock_open
    opened = False

    with mockKrakenXDevice.connect() as cm:
        assert cm == mockKrakenXDevice
        assert opened


def test_kraken_initialize(mockKrakenXDevice):
    (fw_ver,) = mockKrakenXDevice.initialize()

    assert fw_ver[1] == '6.2'


def test_old_kraken_initialize(mockOldKrakenXDevice):
    (fw_ver,) = mockOldKrakenXDevice.initialize()

    assert fw_ver[1] == '2.1.8'


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True)])
def test_kraken_get_status_directly(mockKrakenXDevice, has_hwmon, direct_access):
    if has_hwmon:
        mockKrakenXDevice._hwmon = HwmonDevice(None, None)

    got = mockKrakenXDevice.get_status(direct_access=direct_access)

    expected = [
        ('Liquid temperature', pytest.approx(30.9), '°C'),
        ('Fan speed', 1499, 'rpm'),
        ('Pump speed', 2702, 'rpm'),
    ]

    assert sorted(got) == sorted(expected)


def test_kraken_get_status_from_hwmon(mockKrakenXDevice, tmp_path):
    mockKrakenXDevice._hwmon = HwmonDevice('mock_module', tmp_path)
    (tmp_path / 'temp1_input').write_text('20900\n')
    (tmp_path / 'fan1_input').write_text('2499\n')
    (tmp_path / 'fan2_input').write_text('1702\n')

    got = mockKrakenXDevice.get_status()

    expected = [
        ('Liquid temperature', pytest.approx(20.9), '°C'),
        ('Fan speed', 2499, 'rpm'),
        ('Pump speed', 1702, 'rpm'),
    ]

    assert sorted(got) == sorted(expected)


def test_kraken_not_totally_broken(mockKrakenXDevice):
    """Reasonable example calls to untested APIs do not raise exceptions."""
    dev = mockKrakenXDevice

    dev.initialize()
    dev.set_color(channel='ring', mode='loading', colors=iter([[90, 80, 0]]),
                  speed='slowest')
    dev.set_speed_profile(channel='fan',
                          profile=iter([(20, 20), (30, 40), (40, 100)]))
    dev.set_fixed_speed(channel='pump', duty=50)
    dev.set_instantaneous_speed(channel='pump', duty=50)


def test_kraken_set_fixed_speeds(mockOldKrakenXDevice):
    mockOldKrakenXDevice.set_fixed_speed(channel='fan', duty=42)
    mockOldKrakenXDevice.set_fixed_speed(channel='pump', duty=84)

    fan_report, pump_report = mockOldKrakenXDevice.device.sent

    assert fan_report.number == 2
    assert fan_report.data[0:4] == [0x4d, 0, 0, 42]
    assert pump_report.number == 2
    assert pump_report.data[0:4] == [0x4d, 0x40, 0, 84]


def test_kraken_speed_profiles_not_supported(mockOldKrakenXDevice):

    with pytest.raises(NotSupportedByDevice):
        mockOldKrakenXDevice.set_speed_profile('fan', [(20, 42)])

    with pytest.raises(NotSupportedByDevice):
        mockOldKrakenXDevice.set_speed_profile('pump', [(20, 84)])


def test_krakenM_initialize(mockKrakenMDevice):
    (fw_ver,) = mockKrakenMDevice.initialize()
    assert fw_ver[1] == '6.2'


def test_krakenM_get_status(mockKrakenMDevice):
    assert mockKrakenMDevice.get_status() == []


def test_krakenM_speed_control_not_supported(mockKrakenMDevice):
    with pytest.raises(NotSupportedByDevice):
        mockKrakenMDevice.set_fixed_speed('fan', 42)

    with pytest.raises(NotSupportedByDevice):
        mockKrakenMDevice.set_fixed_speed('pump', 84)

    with pytest.raises(NotSupportedByDevice):
        mockKrakenMDevice.set_speed_profile('fan', [(20, 42)])

    with pytest.raises(NotSupportedByDevice):
        mockKrakenMDevice.set_speed_profile('pump', [(20, 84)])


def test_krakenM_not_totally_broken(mockKrakenMDevice):
    """Reasonable example calls to untested APIs do not raise exceptions."""
    dev = mockKrakenMDevice
    dev.initialize()
    dev.set_color(channel='ring', mode='loading', colors=iter([[90, 80, 0]]),
                  speed='slowest')



import pytest
from _testutils import MockHidapiDevice, Report
from pytest import approx

from datetime import timedelta

from liquidctl.driver.corsair_hid_psu import CorsairHidPsu, OCPMode, FanControlMode
from liquidctl.driver.hwmon import HwmonDevice


# https://github.com/liquidctl/liquidctl/issues/300#issuecomment-788302513
SAMPLE_PAGED_RESPONSES = [
    [
        '038bffd2',
        '038c2bf0',
        '03963e08',
    ],
    [
        '038b41d1',
        '038c1be0',
        '039610f8',
    ],
    [
        '038bd3d0',
        '038c09e0',
        '039603f8',
    ],
]

SAMPLE_RESPONSES = [
    # https://github.com/liquidctl/liquidctl/issues/300#issuecomment-788302513
    '033b1b',
    '034013d1',
    '03441ad2',
    '034680e2',
    '034f46',
    '0388ccf9',
    '038d86f0',
    '038e6af0',
    '0399434f5253414952',
    '039a524d3130303069',
    '03d46d9febfe',
    '03d802',
    '03ee4608',
    'fe03524d3130303069',

    # https://github.com/liquidctl/liquidctl/pull/54#issuecomment-543760522
    '03d29215',
    '03d1224711',

    # artificial
    '0390c803',
    '03f001',
]


class MockPsu(MockHidapiDevice):
    def __init__(self, *args, **kwargs):
        self._page = 0;
        super().__init__(*args, **kwargs)

    def write(self, data):
        super().write(data)
        data = data[1:]  # skip unused report ID

        reply = bytearray(64)

        if data[0] == 2 and data[1] == 0:
            self._page = data[2]
            reply[0:3] = data[0:3]
            self.preload_read(Report(0, reply))
        else:
            cmd = f'{data[1]:02x}'
            samples = [x for x in SAMPLE_PAGED_RESPONSES[self._page] if x[2:4] == cmd]
            if not samples:
                samples = [x for x in SAMPLE_RESPONSES if x[2:4] == cmd]
            if not samples:
                raise KeyError(cmd)
            reply[0] = data[0]
            reply[1:len(data)] = bytes.fromhex(samples[0][2:])
            self.preload_read(Report(0, reply))


@pytest.fixture
def mock_psu():
    pid, vid, desc, kwargs = CorsairHidPsu._MATCHES[0]
    device = MockPsu(vendor_id=vid, product_id=pid, address='addr')
    return CorsairHidPsu(device, f'Mock {desc}', **kwargs)


def test_not_totally_broken(mock_psu):

    mock_psu.set_fixed_speed(channel='fan', duty=50)
    report_id, report_data = mock_psu.device.sent[0]
    assert report_id == 0
    assert len(report_data) == 64


def test_dont_inject_report_ids(mock_psu):

    mock_psu.set_fixed_speed(channel='fan', duty=50)
    report_id, report_data = mock_psu.device.sent[0]
    assert report_id == 0
    assert len(report_data) == 64


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True), (True, False)])
def test_initializes(mock_psu, has_hwmon, direct_access, tmp_path):
    if has_hwmon:
        mock_psu._hwmon = HwmonDevice('mock_module', tmp_path)

    # TODO check the result
    _ = mock_psu.initialize(direct_access=direct_access)

    writes = len(mock_psu.device.sent)
    if not has_hwmon or direct_access:
        assert writes > 0
    else:
        assert writes == 0


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True)])
def test_reads_status_directly(mock_psu, has_hwmon, direct_access):
    if has_hwmon:
        mock_psu._hwmon = HwmonDevice(None, None)

    got = mock_psu.get_status(direct_access=direct_access)

    expected = [
        ('Current uptime', timedelta(seconds=5522), ''),
        ('Total uptime', timedelta(days=13, seconds=9122), ''),
        ('VRM temperature', approx(33.5, rel=1e-3), '°C'),
        ('Case temperature', approx(26.5, rel=1e-3), '°C'),
        ('Fan control mode', FanControlMode.SOFTWARE, ''),
        ('Fan speed', approx(968, rel=1e-3), 'rpm'),
        ('Input voltage', approx(230, rel=1e-3), 'V'),
        ('Total power output', approx(140, rel=1e-3), 'W'),
        ('+12V OCP mode', OCPMode.MULTI_RAIL, ''),
        ('+12V output voltage', approx(11.98, rel=1e-3), 'V'),
        ('+12V output current', approx(10.75, rel=1e-3), 'A'),
        ('+12V output power', approx(124, rel=1e-3), 'W'),
        ('+5V output voltage', approx(5.016, rel=1e-3), 'V'),
        ('+5V output current', approx(1.688, rel=1e-3), 'A'),
        ('+5V output power', approx(8, rel=1e-3), 'W'),
        ('+3.3V output voltage', approx(3.297, rel=1e-3), 'V'),
        ('+3.3V output current', approx(0.562, rel=1e-3), 'A'),
        ('+3.3V output power', approx(1.5, rel=1e-3), 'W'),
        ('Estimated input power', approx(153, abs=1), 'W'),
        ('Estimated efficiency', approx(92, abs=1), '%'),
    ]

    assert sorted(got) == sorted(expected)


def test_reads_status_from_hwmon(mock_psu, tmp_path):
    mock_psu.device.write = None  # make sure we aren't writing to the mock device

    mock_psu._hwmon = HwmonDevice('mock_module', tmp_path)
    (tmp_path / 'temp1_input').write_text('33500\n')
    (tmp_path / 'temp2_input').write_text('26500\n')
    (tmp_path / 'fan1_input').write_text('968\n')
    (tmp_path / 'in0_input').write_text('230000\n')
    (tmp_path / 'power1_input').write_text('140000000\n')
    (tmp_path / 'in1_input').write_text('11980\n')
    (tmp_path / 'curr2_input').write_text('10750\n')
    (tmp_path / 'power2_input').write_text('124000000\n')
    (tmp_path / 'in2_input').write_text('5016\n')
    (tmp_path / 'curr3_input').write_text('1688\n')
    (tmp_path / 'power3_input').write_text('8000000\n')
    (tmp_path / 'in3_input').write_text('3297\n')
    (tmp_path / 'curr4_input').write_text('562\n')
    (tmp_path / 'power4_input').write_text('1500000\n')

    got = mock_psu.get_status()

    expected = [
        ('VRM temperature', approx(33.5, rel=1e-3), '°C'),
        ('Case temperature', approx(26.5, rel=1e-3), '°C'),
        ('Fan speed', approx(968, rel=1e-3), 'rpm'),
        ('Input voltage', approx(230, rel=1e-3), 'V'),
        ('Total power output', approx(140, rel=1e-3), 'W'),
        ('+12V output voltage', approx(11.98, rel=1e-3), 'V'),
        ('+12V output current', approx(10.75, rel=1e-3), 'A'),
        ('+12V output power', approx(124, rel=1e-3), 'W'),
        ('+5V output voltage', approx(5.016, rel=1e-3), 'V'),
        ('+5V output current', approx(1.688, rel=1e-3), 'A'),
        ('+5V output power', approx(8, rel=1e-3), 'W'),
        ('+3.3V output voltage', approx(3.297, rel=1e-3), 'V'),
        ('+3.3V output current', approx(0.562, rel=1e-3), 'A'),
        ('+3.3V output power', approx(1.5, rel=1e-3), 'W'),
        ('Estimated input power', approx(153, abs=1), 'W'),
        ('Estimated efficiency', approx(92, abs=1), '%'),
    ]

    assert sorted(got) == sorted(expected)



"""liquidctl – monitor and control liquid coolers and other devices.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style

import sys
from liquidctl.cli import main

if __name__ == "__main__":
    sys.exit(main())



import pytest
from collections import deque

from liquidctl import ExpectationNotMet
from liquidctl.driver.commander_core import CommanderCore
from _testutils import noop
from liquidctl.util import u16le_from


def int_to_le(num, length=2, byteorder='little', signed=False):
    return int(num).to_bytes(length=length, byteorder=byteorder, signed=signed)


class MockCommanderCoreDevice:
    def __init__(self):
        self.vendor_id = 0x1b1c
        self.product_id = 0x0c1c
        self.address = 'addr'
        self.path = b'path'
        self.release_number = None
        self.serial_number = None
        self.bus = None
        self.port = None

        self.open = noop
        self.close = noop
        self.clear_enqueued_reports = noop

        self._read = deque()
        self.sent = list()

        self._last_write = bytes()
        self._modes = {}
        self._awake = False

        self.response_prefix = ()
        self.firmware_version = (0x00, 0x00, 0x00)
        self.led_counts = (None, None, None, None, None, None, None)
        self.speeds_mode = (0, 0, 0, 0, 0, 0, 0)
        self.speeds = (None, None, None, None, None, None, None)
        self.fixed_speeds = (0, 0, 0, 0, 0, 0, 0)
        self.temperatures = (None, None)

    def read(self, length):
        data = bytearray([0x00, self._last_write[2], 0x00])
        data.extend(self.response_prefix)

        if self._last_write[2] == 0x02:  # Firmware version
            for i in range(0, 3):
                data.append(self.firmware_version[i])
        if self._awake:
            if self._last_write[2] == 0x08:  # Get data
                channel = self._last_write[3]
                mode = self._modes.get(channel)
                if mode[1] == 0x00:
                    if mode[0] == 0x17:  # Get speeds
                        data.extend([0x06, 0x00])
                        data.append(len(self.speeds))
                        for i in self.speeds:
                            if i is None:
                                data.extend([0x00, 0x00])
                            else:
                                data.extend(int_to_le(i))
                    elif mode[0] == 0x1a:  # Speed devices connected
                        data.extend([0x09, 0x00])
                        data.append(len(self.speeds))
                        for i in self.speeds:
                            data.extend([0x01 if i is None else 0x07])
                    elif mode[0] == 0x20:  # LED detect
                        data.extend([0x0f, 0x00])
                        data.append(len(self.led_counts))
                        for i in self.led_counts:
                            if i is None:
                                data.extend(int_to_le(3)+int_to_le(0))
                            else:
                                data.extend(int_to_le(2))
                                data.extend(int_to_le(i))
                    elif mode[0] == 0x21:  # Get temperatures
                        data.extend([0x10, 0x00])
                        data.append(len(self.temperatures))
                        for i in self.temperatures:
                            if i is None:
                                data.append(1)
                                data.extend(int_to_le(0))
                            else:
                                data.append(0)
                                data.extend(int_to_le(int(i*10)))
                    else:
                        raise NotImplementedError(f'Read for {mode.hex(":")}')
                elif mode[1] == 0x6d:
                    if mode[0] == 0x60:
                        data.extend([0x03, 0x00])
                        data.append(len(self.speeds_mode))
                        for i in self.speeds_mode:
                            data.append(i)
                    elif mode[0] == 0x61:
                        data.extend([0x04, 0x00])
                        data.append(len(self.fixed_speeds))
                        for i in self.fixed_speeds:
                            data.extend(int_to_le(i))
                    else:
                        raise NotImplementedError(f'Read for {mode.hex(":")}')
                else:
                    raise NotImplementedError(f'Read for {mode.hex(":")}')

        return list(data)[:length]

    def write(self, data):
        data = bytes(data)  # ensure data is convertible to bytes
        self._last_write = data
        if data[0] != 0x00 or data[1] != 0x08:
            raise ValueError('Start of packets going out should be 00:08')
        if data[2] == 0x0d:
            channel = data[3]
            if  self._modes.get(channel) is None:
                self._modes[channel] = data[4:6]
            else:
                raise ExpectationNotMet('Previous channel was not reset')
        elif data[2] == 0x05 and data[3] == 0x01:
            self._modes[data[4]] = None
        elif data[2] == 0x01 and data[3] == 0x03 and data[4] == 0x00:
            self._awake = data[5] == 0x02
        elif self._awake:
            if data[2] == 0x06:  # Write command
                channel = data[3]
                mode = self._modes.get(channel)
                length = u16le_from(data[4:6])
                data_type = data[8:10]
                written_data = data[10:8+length]
                if mode[1] == 0x6d:
                    if mode[0] == 0x60 and list(data_type) == [0x03, 0x00]:
                        self.speeds_mode = tuple(written_data[i+1] for i in range(0, written_data[0]))
                    elif mode[0] == 0x61 and list(data_type) == [0x04, 0x00]:
                        self.fixed_speeds = tuple(u16le_from(written_data[i*2+1:i*2+3]) for i in range(0, written_data[0]))
                    else:
                        raise NotImplementedError('Invalid Write command')
                else:
                    raise NotImplementedError('Invalid Write command')

        return len(data)


@pytest.fixture
def commander_core_device():
    device = MockCommanderCoreDevice()
    core = CommanderCore(device, 'Corsair Commander Core (experimental)', True)
    core.connect()
    return core


def test_initialize_commander_core(commander_core_device):
    commander_core_device.device.firmware_version = (0x01, 0x02, 0x21)
    commander_core_device.device.speeds = (None, 104, None, None, None, None, 918)
    commander_core_device.device.led_counts = (27, None, 1, 2, 4, 8, 16)
    commander_core_device.device.temperatures = (None, 45.6)
    res = commander_core_device.initialize()

    assert len(res) == 17

    assert res[0][1] == '1.2.33'  # Firmware

    # LED counts
    assert res[1][1] == 27
    assert res[2][1] is None
    assert res[3][1] == 1
    assert res[4][1] == 2
    assert res[5][1] == 4
    assert res[6][1] == 8
    assert res[7][1] == 16

    # Speed devices connected
    assert not res[8][1]
    assert res[9][1]
    assert not res[10][1]
    assert not res[11][1]
    assert not res[12][1]
    assert not res[13][1]
    assert res[14][1]

    # Temperature sensors connected
    assert not res[15][1]
    assert res[16][1]

    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_initialize_error_commander_core(commander_core_device):
    """This tests sends invalid data to ensure the device gets set back to hardware mode on error"""
    commander_core_device.device.response_prefix = (0x00, 0x00)

    with pytest.raises(ExpectationNotMet):
        commander_core_device.initialize()

    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_status_commander_core(commander_core_device):
    commander_core_device.device.speeds = (2357, 918, 903, 501, 1104, 1824, 104)
    commander_core_device.device.temperatures = (12.3, 45.6)
    res = commander_core_device.get_status()

    assert len(res) == 9

    # Speeds of pump and fans
    assert res[0][1] == 2357
    assert res[1][1] == 918
    assert res[2][1] == 903
    assert res[3][1] == 501
    assert res[4][1] == 1104
    assert res[5][1] == 1824
    assert res[6][1] == 104

    # Temperatures
    assert res[7][1] == 12.3
    assert res[8][1] == 45.6

    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_status_error_commander_core(commander_core_device):
    """This tests sends invalid data to ensure the device gets set back to hardware mode on error"""
    commander_core_device.device.response_prefix = (0x00, 0x00)

    with pytest.raises(ExpectationNotMet):
        commander_core_device.initialize()

    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_set_fixed_speed_fan2_commander_core(commander_core_device):
    """This tests setting the speed of a single channel"""
    commander_core_device.device.speeds_mode = (1, 2, 3, 4, 5, 6, 7)
    commander_core_device.device.fixed_speeds = (8, 9, 10, 11, 12, 13, 14)

    commander_core_device.set_fixed_speed('fan2', 95)

    assert commander_core_device.device.speeds_mode == (1, 2, 0, 4, 5, 6, 7)
    assert commander_core_device.device.fixed_speeds == (8, 9, 95, 11, 12, 13, 14)
    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_set_fixed_speed_fans_commander_core(commander_core_device):
    """This tests setting the speed of all the fans"""
    commander_core_device.device.speeds_mode = (1, 2, 3, 4, 5, 6, 7)
    commander_core_device.device.fixed_speeds = (8, 9, 10, 11, 12, 13, 14)

    commander_core_device.set_fixed_speed('fans', 61)

    assert commander_core_device.device.speeds_mode == (1, 0, 0, 0, 0, 0, 0)
    assert commander_core_device.device.fixed_speeds == (8, 61, 61, 61, 61, 61, 61)
    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_set_fixed_speed_error_commander_core(commander_core_device):
    """This tests sends invalid data to ensure the device gets set back to hardware mode on error"""
    commander_core_device.device.response_prefix = (0x00, 0x00)

    with pytest.raises(ExpectationNotMet):
        commander_core_device.set_fixed_speed('fan1', 95)

    # Ensure device is asleep at end
    assert not commander_core_device.device._awake


def test_parse_channels_commander_core():
    """This test will go through and thoroughly test CommanderCore._parse_channels so we don't have to in other tests"""
    core = CommanderCore(MockCommanderCoreDevice(), 'Corsair Commander Core (experimental)', True)
    tests = [
        ('pump', [0]), ('fans', [1, 2, 3, 4, 5, 6]),
        ('fan1', [1]), ('fan2', [2]), ('fan3', [3]), ('fan4', [4]), ('fan5', [5]), ('fan6', [6])
    ]

    for (val, answer) in tests:
        assert list(core._parse_channels(val)) == answer



def test_parse_channels_commander_core_xt():
    """Test Core XT-specific channel layout"""
    corext = CommanderCore(MockCommanderCoreDevice(), 'Corsair Commander Core (experimental)', False)
    tests = [
        ('fans', [0, 1, 2, 3, 4, 5]),
        ('fan1', [0]), ('fan2', [1]), ('fan3', [2]), ('fan4', [3]), ('fan5', [4]), ('fan6', [5])
    ]

    for (val, answer) in tests:
        assert list(corext._parse_channels(val)) == answer


def test_parse_channels_error_commander_core():
    """This tests to make sure we get an error with an invalid channel"""
    core = CommanderCore(MockCommanderCoreDevice(), 'Corsair Commander Core (experimental)', True)
    with pytest.raises(ValueError):
        core._parse_channels('fan')



import pytest
from _testutils import VirtualBusDevice


def test_connects_and_disconnects_with_context_manager():
    dev = VirtualBusDevice()

    with pytest.raises(RuntimeError):
        with dev.connect():
            assert dev.connected
            raise RuntimeError()

    assert not dev.connected


def test_entering_the_runtime_context_does_not_call_connect():
    dev = VirtualBusDevice()

    with dev.connect(marker=True):
        # since __enter__ takes no arguments, if __enter__ calls connect it
        # will override dev.kwargs['connect'] with {}
        assert 'marker' in dev.call_args['connect'].kwargs


def test_modified_readme_example(capsys):
    from liquidctl import find_liquidctl_devices

    # Find all connected and supported devices.
    devices = find_liquidctl_devices(bus='virtual')  # readme: remove `bus` argument

    for dev in devices:

        # Connect to the device. In this example we use a context manager, but
        # the connection can also be manually managed. The context manager
        # automatically calls `disconnect`; when managing the connection
        # manually, `disconnect` must eventually be called, even if an
        # exception is raised.
        with dev.connect():
            print(f'{dev.description} at {dev.bus}:{dev.address}:')

            # Devices should be initialized after every boot. In this example
            # we assume that this has not been done before.
            print('- initialize')
            init_status = dev.initialize()

            # Print all data returned by `initialize`.
            if init_status:
                for key, value, unit in init_status:
                    print(f'- {key}: {value} {unit}')

            # Get regular status information from the device.
            status = dev.get_status()

            # Print all data returned by `get_status`.
            print('- get status')
            for key, value, unit in status:
                print(f'- {key}: {value} {unit}')

            # For a particular device, set the pump LEDs to red.
            if 'Virtual Bus Device' in dev.description: # readme: replace with 'Kraken'
                print('- set pump to radical red')
                radical_red = [0xff, 0x35, 0x5e]
                dev.set_color(channel='pump', mode='fixed', colors=[radical_red])


    # end of modified example; check that it more or less did what it should

    out, _ = capsys.readouterr()
    assert 'Virtual Bus Device (experimental) at virtual:virtual_address:' in out
    assert 'initialize' in out
    assert 'Firmware version: 3.14.16' in out
    assert 'get status' in out
    assert 'Temperature: 30.4 °C' in out
    assert 'set pump to radical red' in out



import pytest
from _testutils import MockHidapiDevice, Report

from collections import deque

from liquidctl.driver.aura_led import AuraLed

# Sample data for Aura LED controller from ASUS ProArt Z690-Creator WiFi
_INIT_19AF_FIRMWARE_DATA = bytes.fromhex(
    "ec0241554c41332d415233322d30323037000000000000000000000000000000"
    "000000000000000000000000000000000000000000000000000000000000000000"
)
_INIT_19AF_FIRMWARE = Report(_INIT_19AF_FIRMWARE_DATA[0], _INIT_19AF_FIRMWARE_DATA[1:])

_INIT_19AF_CONFIG_DATA = bytes.fromhex(
    "ec3000001e9f03010000783c00010000783c00010000783c0000000000000001"
    "040201f40000000000000000000000000000000000000000000000000000000000"
)
_INIT_19AF_CONFIG = Report(_INIT_19AF_CONFIG_DATA[0], _INIT_19AF_CONFIG_DATA[1:])


@pytest.fixture
def mockAuraLed_19AFDevice():
    device = MockHidapiDevice(vendor_id=0x0B05, product_id=0x19AF, address="addr")
    dev = AuraLed(device, "mock Aura LED Controller")
    dev.connect()
    return dev


def test_aura_led_19AF_device_command_format(mockAuraLed_19AFDevice):
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_FIRMWARE)
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_CONFIG)
    mockAuraLed_19AFDevice.initialize()  # should perform 3 writes
    mockAuraLed_19AFDevice.set_color(
        channel="sync", mode="off", colors=[]
    )  # should perform 14 writes
    assert len(mockAuraLed_19AFDevice.device.sent) == 2 + 14
    for i, (report, data) in enumerate(mockAuraLed_19AFDevice.device.sent):
        assert report == 0xEC
        assert len(data) == 64


def test_aura_led_19AF_device_get_status(mockAuraLed_19AFDevice):
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_CONFIG)
    assert mockAuraLed_19AFDevice.get_status() != []


def test_aura_led_19AF_device_initialize_status(mockAuraLed_19AFDevice):
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_FIRMWARE)
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_CONFIG)
    status_list = mockAuraLed_19AFDevice.initialize()
    firmware_tuple = status_list[0]
    assert firmware_tuple[1] == "AULA3-AR32-0207"


def test_aura_led_19AF_device_off_with_some_channel(mockAuraLed_19AFDevice):
    colors = [[0xFF, 0, 0x80]]  # should be ignored
    mockAuraLed_19AFDevice.set_color(channel="led2", mode="off", colors=iter(colors))
    assert len(mockAuraLed_19AFDevice.device.sent) == 5
    data1 = mockAuraLed_19AFDevice.device.sent[0].data
    data2 = mockAuraLed_19AFDevice.device.sent[1].data
    assert data1[1] == 0x01  # key for led2
    assert data1[4] == 0x00  # off
    assert data2[2] == 0x02  # channel led2
    assert data2[7:10] == [0x00, 0x00, 0x00]


def test_aura_led_19AF_static_with_some_channel(mockAuraLed_19AFDevice):
    colors = [[0xFF, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockAuraLed_19AFDevice.set_color(channel="led2", mode="static", colors=iter(colors))
    assert len(mockAuraLed_19AFDevice.device.sent) == 5
    data1 = mockAuraLed_19AFDevice.device.sent[0].data
    data2 = mockAuraLed_19AFDevice.device.sent[1].data
    assert data1[1] == 0x01  # key for led2
    assert data1[4] == 0x01  # static mode
    assert data2[2] == 0x02  # channel led2
    assert data2[7:10] == [0xFF, 0x00, 0x80]


def test_aura_led_19AF_spectrum_cycle_with_some_channel(mockAuraLed_19AFDevice):
    colors = [[0xFF, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockAuraLed_19AFDevice.set_color(channel="led3", mode="spectrum_cycle", colors=iter(colors))
    assert len(mockAuraLed_19AFDevice.device.sent) == 5
    data1 = mockAuraLed_19AFDevice.device.sent[0].data
    data2 = mockAuraLed_19AFDevice.device.sent[1].data
    assert data1[1] == 0x01  # key for led3
    assert data1[4] == 0x04  # spectrum cycle
    assert data2[2] == 0x04  # channel led3
    assert data2[7:10] == [0x00, 0x00, 0x00]


def test_aura_led_19AF_device_sync_channel(mockAuraLed_19AFDevice):
    colors = [[0xFF, 0, 0x80]]
    mockAuraLed_19AFDevice.set_color(channel="sync", mode="static", colors=iter(colors))
    assert len(mockAuraLed_19AFDevice.device.sent) == 14  # 14 writes


def test_aura_led_19AF_device_invalid_set_color_arguments(mockAuraLed_19AFDevice):

    with pytest.raises(KeyError):
        mockAuraLed_19AFDevice.set_color("invalid", "off", [])

    with pytest.raises(KeyError):
        mockAuraLed_19AFDevice.set_color("led2", "invalid", [])

    with pytest.raises(ValueError):
        mockAuraLed_19AFDevice.set_color("led3", "static", [])


def test_aura_led_19AF_device_initialize_status(mockAuraLed_19AFDevice):
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_FIRMWARE)
    mockAuraLed_19AFDevice.device.preload_read(_INIT_19AF_CONFIG)
    status_list = mockAuraLed_19AFDevice.initialize()
    firmware_tuple = status_list[0]
    assert firmware_tuple[1] == "AULA3-AR32-0207"



# uses the psf/black style

import pytest

from liquidctl.driver.hwmon import HwmonDevice


@pytest.fixture
def mock_hwmon(tmp_path):
    hwmon = tmp_path / "hwmon7"
    hwmon.mkdir()

    (hwmon / "fan1_input").write_text("1499\n")
    (hwmon / "fan1_label").write_text("Pump Speed\n")

    return HwmonDevice("mock_module", hwmon)


def test_has_module(mock_hwmon):
    assert mock_hwmon.driver == "mock_module"


def test_has_path(mock_hwmon):
    assert mock_hwmon.path.is_dir()


def test_has_name(mock_hwmon):
    assert mock_hwmon.name == "hwmon7"


def test_checks_existing_attribute(mock_hwmon):
    assert mock_hwmon.has_attribute("fan1_input")


def test_checks_non_existing_attribute(mock_hwmon):
    assert not mock_hwmon.has_attribute("bubble1_input")


def test_gets_string(mock_hwmon):
    assert mock_hwmon.get_string("fan1_label") == "Pump Speed"


def test_gets_int(mock_hwmon):
    assert mock_hwmon.read_int("fan1_input") == 1499



"""liquidctl drivers for NVIDIA graphics cards.

Copyright Jonas Malaco, Marshall Asch and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
from enum import unique

from liquidctl.driver.smbus import SmbusDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import RelaxedNamesEnum, check_unsafe

_LOGGER = logging.getLogger(__name__)

# sources for pci device and subsystem device ids:
# - https://www.nv-drivers.eu/nvidia-all-devices.html
# - https://pci-ids.ucw.cz/pci.ids
# - https://gitlab.com/CalcProgrammer1/OpenRGB/-/blob/master/pci_ids/pci_ids.h

# vendor, devices
# FWUPD_GUID = [vendor]:[device] - use hwinfo to inspect
NVIDIA = 0x10de
NVIDIA_GTX_1050 = 0x1c81
NVIDIA_GTX_1050_TI = 0x1c82
NVIDIA_GTX_1060 = 0x1c03
NVIDIA_GTX_1070 = 0x1b81
NVIDIA_GTX_1070_TI = 0x1b82
NVIDIA_GTX_1080 = 0x1b80
NVIDIA_GTX_1080_TI = 0x1b06
NVIDIA_GTX_1650 = 0x1f82
NVIDIA_GTX_1650S = 0x2187
NVIDIA_GTX_1660 = 0x2184
NVIDIA_GTX_1660S = 0x21c4
NVIDIA_GTX_1660_TI = 0x2182
NVIDIA_RTX_2060S = 0x1f47
NVIDIA_RTX_2060S_OC = 0x1f06
NVIDIA_RTX_2060_TU104 = 0x1e89
NVIDIA_RTX_2060_TU106 = 0x1f08
NVIDIA_RTX_2070 = 0x1f02
NVIDIA_RTX_2070S = 0x1ec7
NVIDIA_RTX_2070S_OC = 0x1e84
NVIDIA_RTX_2070_OC = 0x1f07
NVIDIA_RTX_2080 = 0x1e82
NVIDIA_RTX_2080S = 0x1e81
NVIDIA_RTX_2080_REV_A = 0x1e87
NVIDIA_RTX_2080_TI = 0x1e04
NVIDIA_RTX_2080_TI_REV_A = 0x1e07
NVIDIA_RTX_3050 = 0x2507
NVIDIA_RTX_3060 = 0x2503
NVIDIA_RTX_3060_GA104 = 0x2487
NVIDIA_RTX_3060_LHR = 0x2504
NVIDIA_RTX_3060_TI = 0x2486
NVIDIA_RTX_3060_TI_LHR = 0x2489
NVIDIA_RTX_3070 = 0x2484
NVIDIA_RTX_3070_LHR = 0x2488
NVIDIA_RTX_3070_TI = 0x2482
NVIDIA_RTX_3080 = 0x2206
NVIDIA_RTX_3080_12G_LHR = 0x220a
NVIDIA_RTX_3080_LHR = 0x2216
NVIDIA_RTX_3080_TI = 0x2208
NVIDIA_RTX_3090 = 0x2204

# subsystem vendor ASUS, subsystem devices
# PCI_SUBSYS_ID = [subsystem vendor]:[subsystem device] - use hwinfo to inspect
ASUS = 0x1043
ASUS_STRIX_GTX_1050_O2G = 0x85d8
ASUS_STRIX_GTX_1050_TI_O4G = 0x85cd
ASUS_STRIX_GTX_1050_TI_O4G_2 = 0x85d1
ASUS_STRIX_GTX_1060_6G = 0x85a4
ASUS_STRIX_GTX_1060_O6G = 0x85ac
ASUS_STRIX_GTX_1070 = 0x8598
ASUS_STRIX_GTX_1070_OC = 0x8599
ASUS_STRIX_GTX_1070_TI_8G = 0x861d
ASUS_STRIX_GTX_1070_TI_A8G = 0x861e
ASUS_STRIX_GTX_1080 = 0x8592
ASUS_STRIX_GTX_1080_A8G = 0x85aa
ASUS_STRIX_GTX_1080_O8G = 0x85f9
ASUS_STRIX_GTX_1080_TI = 0x85eb
ASUS_STRIX_GTX_1080_TI_11G = 0x85f1
ASUS_STRIX_GTX_1080_TI_O11G = 0x85ea
ASUS_STRIX_GTX_1080_TI_O11G_A02 = 0x85e4
ASUS_STRIX_GTX_1650S_OC = 0x874f
ASUS_STRIX_GTX_1660S_O6G = 0x8752
ASUS_STRIX_GTX_1660_TI_OC = 0x86a5
ASUS_STRIX_RTX_2060S_8G = 0x8730
ASUS_STRIX_RTX_2060S_A8G = 0x86fc
ASUS_STRIX_RTX_2060S_A8G_EVO = 0x8703
ASUS_STRIX_RTX_2060S_O8G = 0x86fb
ASUS_STRIX_RTX_2060_EVO = 0x86d3
ASUS_STRIX_RTX_2060_O6G = 0x868e
ASUS_STRIX_RTX_2060_O6G_EVO = 0x8775
ASUS_STRIX_RTX_2070S_8G_8707 = 0x8707
ASUS_STRIX_RTX_2070S_A8G = 0x8728
ASUS_STRIX_RTX_2070S_A8G_86FF = 0x86ff
ASUS_STRIX_RTX_2070S_A8G_8706 = 0x8706
ASUS_STRIX_RTX_2070S_O8G = 0x8727
ASUS_STRIX_RTX_2070S_O8G_8729 = 0x8729
ASUS_STRIX_RTX_2070_A8G = 0x8671
ASUS_STRIX_RTX_2070_O8G = 0x8670
ASUS_STRIX_RTX_2080S_A8G = 0x8712
ASUS_STRIX_RTX_2080S_O8G = 0x8711
ASUS_STRIX_RTX_2080_O8G = 0x865f
ASUS_STRIX_RTX_2080_TI_11G = 0x8687
ASUS_STRIX_RTX_2080_TI_OC = 0x866a
ASUS_TUF_RTX_3060_TI_O8G_OC = 0x87c6

# subsystem vendor EVGA, subsystem devices
# PCI_SUBSYS_ID = [subsystem vendor]:[subsystem device] - use hwinfo to inspect
EVGA = 0x3842
EVGA_GTX_1070_FTW = 0x6276
EVGA_GTX_1070_FTW_DT_GAMING = 0x6274
EVGA_GTX_1070_FTW_HYBRID = 0x6278
EVGA_GTX_1070_TI_FTW2 = 0x6775
EVGA_GTX_1080_FTW = 0x6286


@unique
class _ModeEnum(bytes, RelaxedNamesEnum):
    def __new__(cls, value, required_colors):
        obj = bytes.__new__(cls, [value])
        obj._value_ = value
        obj.required_colors = required_colors
        return obj

    def __str__(self):
        return self.name.capitalize()


class _NvidiaI2CDriver():
    """Generic NVIDIA I²C driver."""

    _VENDOR = None
    _ADDRESSES = []
    _MATCHES = []

    @classmethod
    def pre_probe(cls, smbus, vendor=None, product=None, address=None, match=None,
              release=None, serial=None, **kwargs):

        if (vendor and vendor != cls._VENDOR) \
                or (address and int(address, base=16) not in cls._ADDRESSES) \
                or release or serial:  # filters can never match, always None
            return

        if smbus.parent_subsystem_vendor != cls._VENDOR \
                or smbus.parent_vendor != NVIDIA \
                or smbus.parent_driver != 'nvidia':
            return

        for dev_id, sub_dev_id, desc in cls._MATCHES:
            if (product and product != sub_dev_id) \
                    or (match and match.lower() not in desc.lower()):
                continue

            if smbus.parent_subsystem_device != sub_dev_id \
                    or smbus.parent_device != dev_id \
                    or not smbus.description.startswith('NVIDIA i2c adapter 1 '):
                continue

            yield (dev_id, sub_dev_id, desc)


class EvgaPascal(SmbusDriver, _NvidiaI2CDriver):
    """NVIDIA series 10 (Pascal) graphics card from EVGA."""

    _REG_MODE = 0x0c
    _REG_RED = 0x09
    _REG_GREEN = 0x0a
    _REG_BLUE = 0x0b
    _REG_PERSIST = 0x23
    _PERSIST = 0xe5

    _VENDOR = EVGA
    _ADDRESSES = [0x49]
    _MATCHES = [
        (NVIDIA_GTX_1070, EVGA_GTX_1070_FTW, 'EVGA GTX 1070 FTW'),
        (NVIDIA_GTX_1070, EVGA_GTX_1070_FTW_DT_GAMING, 'EVGA GTX 1070 FTW DT Gaming'),
        (NVIDIA_GTX_1070, EVGA_GTX_1070_FTW_HYBRID, 'EVGA GTX 1070 FTW Hybrid'),
        (NVIDIA_GTX_1070_TI, EVGA_GTX_1070_TI_FTW2, 'EVGA GTX 1070 Ti FTW2'),
        (NVIDIA_GTX_1080, EVGA_GTX_1080_FTW, 'EVGA GTX 1080 FTW'),
    ]

    @unique
    class Mode(_ModeEnum):
        OFF = (0x00, 0)
        FIXED = (0x01, 1)
        RAINBOW = (0x02, 0)
        BREATHING = (0x05, 1)

    @classmethod
    def probe(cls, smbus, vendor=None, product=None, address=None, match=None,
              release=None, serial=None, **kwargs):

        assert len(cls._ADDRESSES) == 1, 'unexpected extra address candidates'

        pre_probed = super().pre_probe(smbus, vendor, product, address, match,
                                        release, serial, **kwargs)


        for dev_id, sub_dev_id, desc in pre_probed:
            dev = cls(smbus, desc, vendor_id=EVGA, product_id=EVGA_GTX_1080_FTW,
                      address=cls._ADDRESSES[0])
            _LOGGER.debug('%s identified: %s', cls.__name__, desc)
            yield dev

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'experimental' in self.description:
            self._UNSAFE = ['smbus', 'experimental_evga_gpu']
        else:
            self._UNSAFE = ['smbus']

    def get_status(self, verbose=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        # only RGB lighting information can be fetched for now; as that isn't
        # super interesting, only enable it in verbose mode

        if not verbose:
            return []

        if not check_unsafe(*self._UNSAFE, **kwargs):
            _LOGGER.warning("%s: disabled, requires unsafe features '%s'",
                            self.description, ','.join(self._UNSAFE))
            return []

        mode = self.Mode(self._smbus.read_byte_data(self._address, self._REG_MODE))
        status = [('Mode', mode, '')]

        if mode.required_colors > 0:
            r = self._smbus.read_byte_data(self._address, self._REG_RED)
            g = self._smbus.read_byte_data(self._address, self._REG_GREEN)
            b = self._smbus.read_byte_data(self._address, self._REG_BLUE)
            status.append(('Color', f'{r:02x}{g:02x}{b:02x}', ''))

        return status

    def set_color(self, channel, mode, colors, non_volatile=False, **kwargs):
        """Set the RGB lighting mode and, when applicable, color.

        The table bellow summarizes the available channels, modes and their
        associated number of required colors.

        | Channel  | Mode      | Required colors |
        | -------- | --------- | --------------- |
        | led      | off       |               0 |
        | led      | fixed     |               1 |
        | led      | breathing |               1 |
        | led      | rainbow   |               0 |

        The settings configured on the device are normally volatile, and are
        cleared whenever the graphics card is powered down (OS and UEFI power
        saving settings can affect when this happens).

        It is possible to store them in non-volatile controller memory by
        passing `non_volatile=True`.  But as this memory has some unknown yet
        limited maximum number of write cycles, volatile settings are
        preferable, if the use case allows for them.
        """

        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        colors = list(colors)

        try:
            mode = self.Mode[mode]
        except KeyError:
            raise ValueError(f'invalid mode: {mode!r}') from None

        if len(colors) < mode.required_colors:
            raise ValueError(f'{mode} mode requires {mode.required_colors} colors')

        if len(colors) > mode.required_colors:
            _LOGGER.debug('too many colors, dropping to %d', mode.required_colors)
            colors = colors[:mode.required_colors]

        self._smbus.write_byte_data(self._address, self._REG_MODE, mode.value)

        for r, g, b in colors:
            self._smbus.write_byte_data(self._address, self._REG_RED, r)
            self._smbus.write_byte_data(self._address, self._REG_GREEN, g)
            self._smbus.write_byte_data(self._address, self._REG_BLUE, b)

        if non_volatile:
            # the following write always fails, but nonetheless induces persistence
            try:
                self._smbus.write_byte_data(self._address, self._REG_PERSIST, self._PERSIST)
            except OSError as err:
                _LOGGER.debug('expected OSError when writing to _REG_PERSIST: %s', err)

    def initialize(self, **kwargs):
        """Initialize the device."""
        pass

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


class RogTuring(SmbusDriver, _NvidiaI2CDriver):
    """NVIDIA series 10 (Pascal) or 20 (Turing) graphics card from ASUS."""

    _REG_RED = 0x04
    _REG_GREEN = 0x05
    _REG_BLUE = 0x06
    _REG_MODE = 0x07
    _REG_APPLY = 0x0e
    _SYNC_REG = 0x0c  # unused

    _VENDOR = ASUS
    _ADDRESSES = [0x29, 0x2a, 0x60]
    _MATCHES = [
        # description normalization rules:
        # - remove redundant ROG and meaningless GAMING;
        # - remove redundancies within a particular GPU core;
        # - keep OC when appropriate;
        # - uppercase ASUS and TUF, titlecase Evo, Ti and Stix;
        # - enforce ASUS Strix|TUF GTX|RTX <number> [Super] [Ti] [Evo] [OC] [...] order;
        # - use consistent memory sizes: 2GB, 4GB, 6GB, 8GB (but omit when redundant).
        (NVIDIA_GTX_1050, ASUS_STRIX_GTX_1050_O2G, 'ASUS Strix GTX 1050 OC'),
        (NVIDIA_GTX_1050_TI, ASUS_STRIX_GTX_1050_TI_O4G, 'ASUS Strix GTX 1050 Ti OC'),
        (NVIDIA_GTX_1050_TI, ASUS_STRIX_GTX_1050_TI_O4G_2, 'ASUS Strix GTX 1050 Ti OC'),
        (NVIDIA_GTX_1060, ASUS_STRIX_GTX_1060_6G, 'ASUS Strix GTX 1060 6GB'),
        (NVIDIA_GTX_1060, ASUS_STRIX_GTX_1060_O6G, 'ASUS Strix GTX 1060 OC 6GB'),
        (NVIDIA_GTX_1070, ASUS_STRIX_GTX_1070, 'ASUS Strix GTX 1070'),
        (NVIDIA_GTX_1070, ASUS_STRIX_GTX_1070_OC, 'ASUS Strix GTX 1070 OC'),
        (NVIDIA_GTX_1070_TI, ASUS_STRIX_GTX_1070_TI_8G, 'ASUS Strix GTX 1070 Ti'),
        (NVIDIA_GTX_1070_TI, ASUS_STRIX_GTX_1070_TI_A8G, 'ASUS Strix GTX 1070 Ti Advanced'),
        (NVIDIA_GTX_1080, ASUS_STRIX_GTX_1080, 'ASUS Strix GTX 1080'),
        (NVIDIA_GTX_1080, ASUS_STRIX_GTX_1080_A8G, 'ASUS Strix GTX 1080 Advanced'),
        (NVIDIA_GTX_1080, ASUS_STRIX_GTX_1080_O8G, 'ASUS Strix GTX 1080 OC'),
        (NVIDIA_GTX_1080_TI, ASUS_STRIX_GTX_1080_TI, 'ASUS Strix GTX 1080 Ti'),
        (NVIDIA_GTX_1080_TI, ASUS_STRIX_GTX_1080_TI_11G, 'ASUS Strix GTX 1080 Ti'),
        (NVIDIA_GTX_1080_TI, ASUS_STRIX_GTX_1080_TI_O11G, 'ASUS Strix GTX 1080 Ti OC'),
        (NVIDIA_GTX_1080_TI, ASUS_STRIX_GTX_1080_TI_O11G_A02, 'ASUS Strix GTX 1080 Ti OC'),
        (NVIDIA_GTX_1650S, ASUS_STRIX_GTX_1650S_OC, 'ASUS Strix GTX 1650 Super OC'),
        (NVIDIA_GTX_1660S, ASUS_STRIX_GTX_1660S_O6G, 'ASUS Strix GTX 1660 Super OC'),
        (NVIDIA_GTX_1660_TI, ASUS_STRIX_GTX_1660_TI_OC, 'ASUS Strix GTX 1660 Ti OC'),
        (NVIDIA_RTX_2060S, ASUS_STRIX_RTX_2060S_A8G_EVO, 'ASUS Strix RTX 2060 Super Evo Advanced'),
        (NVIDIA_RTX_2060S_OC, ASUS_STRIX_RTX_2060S_8G, 'ASUS Strix RTX 2060 Super'),
        (NVIDIA_RTX_2060S_OC, ASUS_STRIX_RTX_2060S_A8G, 'ASUS Strix RTX 2060 Super Advanced'),
        (NVIDIA_RTX_2060S_OC, ASUS_STRIX_RTX_2060S_O8G, 'ASUS Strix RTX 2060 Super OC'),
        (NVIDIA_RTX_2060_TU104, ASUS_STRIX_RTX_2060_O6G_EVO, 'ASUS Strix RTX 2060 Evo OC'),
        (NVIDIA_RTX_2060_TU106, ASUS_STRIX_RTX_2060_EVO, 'ASUS Strix RTX 2060 Evo'),
        (NVIDIA_RTX_2060_TU106, ASUS_STRIX_RTX_2060_O6G, 'ASUS Strix RTX 2060 OC'),
        (NVIDIA_RTX_2070S, ASUS_STRIX_RTX_2070S_A8G_86FF, 'ASUS Strix RTX 2070 Super Advanced'),
        (NVIDIA_RTX_2070S_OC, ASUS_STRIX_RTX_2070S_8G_8707, 'ASUS Strix RTX 2070'),
        (NVIDIA_RTX_2070S_OC, ASUS_STRIX_RTX_2070S_A8G, 'ASUS Strix RTX 2070 Super Advanced'),
        (NVIDIA_RTX_2070S_OC, ASUS_STRIX_RTX_2070S_A8G_8706, 'ASUS Strix RTX 2070 Super Advanced'),
        (NVIDIA_RTX_2070S_OC, ASUS_STRIX_RTX_2070S_O8G, 'ASUS Strix RTX 2070 Super OC'),
        (NVIDIA_RTX_2070S_OC, ASUS_STRIX_RTX_2070S_O8G_8729, 'ASUS Strix RTX 2070 Super OC'),
        (NVIDIA_RTX_2070_OC, ASUS_STRIX_RTX_2070_A8G, 'ASUS Strix RTX 2070 Advanced'),
        (NVIDIA_RTX_2070_OC, ASUS_STRIX_RTX_2070_O8G, 'ASUS Strix RTX 2070 OC'),
        (NVIDIA_RTX_2080S, ASUS_STRIX_RTX_2080S_A8G, 'ASUS Strix RTX 2080 Super Advanced'),
        (NVIDIA_RTX_2080S, ASUS_STRIX_RTX_2080S_O8G, 'ASUS Strix RTX 2080 Super OC'),
        (NVIDIA_RTX_2080_REV_A, ASUS_STRIX_RTX_2080_O8G, 'ASUS Strix RTX 2080 OC'),
        (NVIDIA_RTX_2080_TI, ASUS_STRIX_RTX_2080_TI_11G, 'ASUS Strix RTX 2080 Ti'),
        (NVIDIA_RTX_2080_TI_REV_A, ASUS_STRIX_RTX_2080_TI_OC, 'ASUS Strix RTX 2080 Ti OC'),
        (NVIDIA_RTX_3060_TI_LHR, ASUS_TUF_RTX_3060_TI_O8G_OC, 'ASUS TUF RTX 3060 Ti OC'),
    ]

    _SENTINEL_ADDRESS = 0xffff  # intentionally invalid
    _ASUS_GPU_APPLY_VAL = 0x01

    @unique
    class Mode(_ModeEnum):
        OFF = (0x00, 0)  # not a real mode; fixed is sent with RGB = 0
        FIXED = (0x01, 1)
        BREATHING = (0x02, 1)
        FLASH = (0x03, 1)
        RAINBOW = (0x04, 0)

    @classmethod
    def probe(cls, smbus, vendor=None, product=None, address=None, match=None,
              release=None, serial=None, **kwargs):

        ASUS_GPU_MAGIC_VALUE = 0x1589

        pre_probed = super().pre_probe(smbus, vendor, product, address, match,
                                        release, serial, **kwargs)

        for dev_id, sub_dev_id, desc in pre_probed:
            selected_address = None

            if check_unsafe('smbus', **kwargs):
                for address in cls._ADDRESSES:
                    val1 = 0
                    val2 = 0

                    smbus.open()
                    try:
                        val1 = smbus.read_byte_data(address, 0x20)
                        val2 = smbus.read_byte_data(address, 0x21)
                    except:
                        pass
                    smbus.close()

                    if val1 << 8 | val2 == ASUS_GPU_MAGIC_VALUE:
                        selected_address = address
                        break
            else:
                selected_address = cls._SENTINEL_ADDRESS
                _LOGGER.debug('unsafe features not enabled, using sentinel address')

            if selected_address is not None:
                dev = cls(smbus, desc, vendor_id=ASUS, product_id=dev_id,
                          address=selected_address)
                _LOGGER.debug('%s identified: %s at address %02x', cls.__name__,
                              desc, selected_address)
                yield dev

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'experimental' in self.description:
            self._UNSAFE = ['smbus', 'experimental_asus_gpu']
        else:
            self._UNSAFE = ['smbus']

    def get_status(self, verbose=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        # only RGB lighting information can be fetched for now; as that isn't
        # super interesting, only enable it in verbose mode

        if not verbose:
            return []

        if not check_unsafe(*self._UNSAFE, **kwargs):
            _LOGGER.warning("%s: disabled, requires unsafe features '%s'",
                            self.description, ','.join(self._UNSAFE))
            return []

        assert self._address != self._SENTINEL_ADDRESS, \
               'invalid address (probing may not have had access to SMbus)'

        mode = self._smbus.read_byte_data(self._address, self._REG_MODE)
        red = self._smbus.read_byte_data(self._address, self._REG_RED)
        green = self._smbus.read_byte_data(self._address, self._REG_GREEN)
        blue = self._smbus.read_byte_data(self._address, self._REG_BLUE)

        # emulate `OFF` both ways
        if red == green == blue == 0:
            mode = 0

        mode = self.Mode(mode)
        status = [('Mode', mode, '')]

        if mode.required_colors > 0:
            status.append(('Color', f'{red:02x}{green:02x}{blue:02x}', ''))

        return status

    def set_color(self, channel, mode, colors, non_volatile=False, **kwargs):
        """Set the lighting mode, when applicable, color.

        The table bellow summarizes the available channels, modes and their
        associated number of required colors.

        | Channel  | Mode      | Required colors |
        | -------- | --------- | --------------- |
        | led      | off       |               0 |
        | led      | fixed     |               1 |
        | led      | flash     |               1 |
        | led      | breathing |               1 |
        | led      | rainbow   |               0 |

        The settings configured on the device are normally volatile, and are
        cleared whenever the graphics card is powered down (OS and UEFI power
        saving settings can affect when this happens).

        It is possible to store them in non-volatile controller memory by
        passing `non_volatile=True`.  But as this memory has some unknown yet
        limited maximum number of write cycles, volatile settings are
        preferable, if the use case allows for them.

        """

        check_unsafe(*self._UNSAFE, error=True, **kwargs)

        assert self._address != self._SENTINEL_ADDRESS, \
               'invalid address (probing may not have had access to SMbus)'

        colors = list(colors)

        try:
            mode = self.Mode[mode]
        except KeyError:
            raise ValueError(f'invalid mode: {mode!r}') from None

        if len(colors) < mode.required_colors:
            raise ValueError(f'{mode} mode requires {mode.required_colors} colors')

        if len(colors) > mode.required_colors:
            _LOGGER.debug('too many colors, dropping to %d', mode.required_colors)
            colors = colors[:mode.required_colors]

        if mode == self.Mode.OFF:
            self._smbus.write_byte_data(self._address, self._REG_MODE,
                                        self.Mode.FIXED.value)
            self._smbus.write_byte_data(self._address, self._REG_RED, 0x00)
            self._smbus.write_byte_data(self._address, self._REG_GREEN, 0x00)
            self._smbus.write_byte_data(self._address, self._REG_BLUE, 0x00)
        else:
            self._smbus.write_byte_data(self._address, self._REG_MODE, mode.value)
            for r, g, b in colors:
                self._smbus.write_byte_data(self._address, self._REG_RED, r)
                self._smbus.write_byte_data(self._address, self._REG_GREEN, g)
                self._smbus.write_byte_data(self._address, self._REG_BLUE, b)

        if non_volatile:
            self._smbus.write_byte_data(self._address, self._REG_APPLY,
                                        self._ASUS_GPU_APPLY_VAL)

    def initialize(self, **kwargs):
        """Initialize the device."""
        pass

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()



import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.smart_device import SmartDevice

SAMPLE_RESPONSES = [
    '043e00056e00000b5b000301000007200002001e00',
    '04400005b500000b5b000201000007020002001e00',
    '044000053800000b5b000201000007120102001e00',
]


@pytest.fixture
def mockSmartDevice():
    device = MockHidapiDevice(vendor_id=0x1e71, product_id=0x1714, address='addr')
    return SmartDevice(device, 'mock NZXT Smart Device V1', speed_channel_count=3, color_channel_count=1)


# class methods
def test_smart_device_constructor(mockSmartDevice):

    assert mockSmartDevice._speed_channels == {
            'fan1': (0, 0, 100),
            'fan2': (1, 0, 100),
            'fan3': (2, 0, 100),
        }

    assert mockSmartDevice._color_channels == {'led': (0), }


def test_smart_device_not_totally_broken(mockSmartDevice):
    dev = mockSmartDevice

    for i in range(4):
        dev.device.preload_read(Report(0, bytes(63)))

    dev.initialize()
    dev.get_status()

    dev.set_color(channel='led', mode='breathing', colors=iter([[142, 24, 68]]),
                  speed='fastest')

    dev.set_fixed_speed(channel='fan3', duty=50)


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True), (True, False)])
def test_smart_device_initializes(mockSmartDevice, has_hwmon, direct_access, tmp_path):
    dev = mockSmartDevice
    if has_hwmon:
        dev._hwmon = HwmonDevice('mock_module', tmp_path)

    for _, capdata in enumerate(SAMPLE_RESPONSES):
        capdata = bytes.fromhex(capdata)
        dev.device.preload_read(Report(capdata[0], capdata[1:]))

    expected = [
        ('Firmware version', '1.7', ''),
        ('LED accessories', 2, ''),
        ('LED accessory type', 'HUE+ Strip', ''),
        ('LED count (total)', 20, ''),
    ]

    got = dev.initialize(direct_access=direct_access)

    assert expected == got

    writes = len(dev.device.sent)
    if not has_hwmon or direct_access:
        assert writes == 2
    else:
        assert writes == 0


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True)])
def test_smart_device_reads_status_directly(mockSmartDevice, has_hwmon, direct_access):
    dev = mockSmartDevice
    if has_hwmon:
        dev._hwmon = HwmonDevice(None, None)

    for _, capdata in enumerate(SAMPLE_RESPONSES):
        capdata = bytes.fromhex(capdata)
        dev.device.preload_read(Report(capdata[0], capdata[1:]))

    # skip initialize for now, we're not emulating the behavior precisely
    # enough to require it here

    expected = [
        ('Fan 1 speed', 1461, 'rpm'),
        ('Fan 1 voltage', 11.91, 'V'),
        ('Fan 1 current', 0.02, 'A'),
        ('Fan 1 control mode', 'PWM', ''),
        ('Fan 2 speed', 1336, 'rpm'),
        ('Fan 2 voltage', 11.91, 'V'),
        ('Fan 2 current', 0.02, 'A'),
        ('Fan 2 control mode', 'PWM', ''),
        ('Fan 3 speed', 1390, 'rpm'),
        ('Fan 3 voltage', 11.91, 'V'),
        ('Fan 3 current', 0.03, 'A'),
        ('Fan 3 control mode', None, ''),
        ('Noise level', 63, 'dB')
    ]

    got = dev.get_status(direct_access=direct_access)

    assert expected == got


def test_smart_device_reads_status_from_hwmon(mockSmartDevice, tmp_path):
    dev = mockSmartDevice

    dev._hwmon = HwmonDevice('mock_module', tmp_path)
    (tmp_path / 'fan1_input').write_text('1461\n')
    (tmp_path / 'in0_input').write_text('11910\n')
    (tmp_path / 'curr1_input').write_text('20\n')
    (tmp_path / 'pwm1_mode').write_text('1\n')
    (tmp_path / 'fan2_input').write_text('1336\n')
    (tmp_path / 'in1_input').write_text('11910\n')
    (tmp_path / 'curr2_input').write_text('20\n')
    (tmp_path / 'pwm2_mode').write_text('0\n')
    (tmp_path / 'fan3_input').write_text('1390\n')
    (tmp_path / 'in2_input').write_text('11910\n')
    (tmp_path / 'curr3_input').write_text('30\n')
    (tmp_path / 'pwm3_mode').write_text('1\n')

    # skip initialize for now, we're not emulating the behavior precisely
    # enough to require it here

    expected = [
        ('Fan 1 speed', 1461, 'rpm'),
        ('Fan 1 voltage', 11.91, 'V'),
        ('Fan 1 current', 0.02, 'A'),
        ('Fan 1 control mode', 'PWM', ''),
        ('Fan 2 speed', 1336, 'rpm'),
        ('Fan 2 voltage', 11.91, 'V'),
        ('Fan 2 current', 0.02, 'A'),
        ('Fan 2 control mode', 'DC', ''),
        ('Fan 3 speed', 1390, 'rpm'),
        ('Fan 3 voltage', 11.91, 'V'),
        ('Fan 3 current', 0.03, 'A'),
        ('Fan 3 control mode', 'PWM', ''),
    ]

    got = dev.get_status()

    assert expected == got



import os
import pytest
import sys
import time
from multiprocessing import Process
from pathlib import Path

from liquidctl.keyval import RuntimeStorage, _FilesystemBackend


@pytest.fixture
def tmpstore(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')
    prefixes = ['prefix']

    backend = _FilesystemBackend(key_prefixes=prefixes, runtime_dirs=[run_dir])
    return RuntimeStorage(prefixes, backend=backend)


def test_loads_and_stores(tmpstore):
    assert tmpstore.load('key') is None
    assert tmpstore.load('key', default=42) == 42

    tmpstore.store('key', '42')

    assert tmpstore.load('key') == '42'
    assert tmpstore.load('key', of_type=int) is None


def test_updates_with_load_store(tmpstore):
    assert tmpstore.load_store('key', lambda x: x) == (None, None)
    assert tmpstore.load_store('key', lambda x: x, default=42) == (None, 42)
    assert tmpstore.load_store('key', lambda x: str(x)) == (42, '42')
    assert tmpstore.load_store('key', lambda x: x, of_type=int) == ('42', None)


def test_fs_backend_stores_truncate_appropriately(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')

    # use a separate reader to prevent caching from masking issues
    writer = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    reader = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])

    writer.store('key', 42)
    assert reader.load('key') == 42

    writer.store('key', 1)
    assert reader.load('key') == 1

    writer.load_store('key', lambda _: 42)
    assert reader.load('key') == 42

    writer.load_store('key', lambda _: 1)
    assert reader.load('key') == 1


def test_fs_backend_loads_from_fallback_dir(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')
    fb_dir = tmpdir.mkdir('fb_dir')

    fallback = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[fb_dir])
    fallback.store('key', 42)

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir, fb_dir])
    assert store.load('key') == 42

    store.store('key', -1)
    assert store.load('key') == -1
    assert fallback.load('key') == 42, 'fallback location was changed'


def test_fs_backend_handles_values_corupted_with_nulls(tmpdir, caplog):
    run_dir = tmpdir.mkdir('run_dir')
    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])

    store.store('key', 42)
    key_file = Path(run_dir).joinpath('prefix', 'key')
    assert key_file.read_bytes() == b'42', 'unit test is unsound'

    key_file.write_bytes(b'\x00')
    val = store.load('key')

    assert val is None
    assert 'was corrupted' in caplog.text


def test_fs_backend_load_store_returns_old_and_new_values(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])

    assert store.load_store('key', lambda _: 42) == (None, 42)
    assert store.load_store('key', lambda x: x + 1) == (42, 43)


def test_fs_backend_load_store_loads_from_fallback_dir(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')
    fb_dir = tmpdir.mkdir('fb_dir')

    fallback = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[fb_dir])
    fallback.store('key', 42)

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir, fb_dir])
    assert store.load_store('key', lambda x: x + 1) == (42, 43)

    assert fallback.load('key') == 42, 'content in fallback location changed'


def test_fs_backend_load_store_loads_from_fallback_dir_that_is_symlink(tmpdir):
    # should deadlock if there is a problem with the lock type or with the
    # handling of fallback paths that point to the same principal/write
    # directory

    run_dir = tmpdir.mkdir('run_dir')
    fb_dir = os.path.join(run_dir, 'symlink')
    try:
        os.symlink(run_dir, fb_dir, target_is_directory=True)
    except OSError as _:
        if sys.platform == 'win32':
            pytest.skip('unable to create Windows symlink with current permissions')
        else:
            raise

    # don't store any initial value so that the fallback location is checked

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir, fb_dir])
    assert store.load_store('key', lambda x: 42) == (None, 42)

    fallback = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[fb_dir])
    assert fallback.load('key') == 42, 'content in fallback symlink did not change'


def test_fs_backend_load_store_is_atomic(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    store.store('key', 42)

    ps = [
        Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', .2)),
        Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', .2)),
        Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', .2)),
    ]

    start_time = time.monotonic()

    for p in ps:
        p.start()

    for p in ps:
        p.join()

    elapsed = (time.monotonic() - start_time)

    assert store.load('key') == 45
    assert elapsed >= .2 * len(ps)


def test_fs_backend_loads_honor_load_store_locking(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    store.store('key', 42)

    ps = [
        Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', .2)),
        Process(target=_fs_mp_check_key, args=(run_dir, 'prefix', 'key', 43)),
    ]

    ps[0].start()
    time.sleep(.1)
    ps[1].start()

    for p in ps:
        p.join()


def test_fs_backend_stores_honor_load_store_locking(tmpdir):
    run_dir = tmpdir.mkdir('run_dir')

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    store.store('key', 42)

    ps = [
        Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', .2)),
        Process(target=_fs_mp_store_key, args=(run_dir, 'prefix', 'key', -1)),
    ]

    start_time = time.monotonic()

    ps[0].start()
    time.sleep(.1)
    ps[1].start()

    # join second process first
    ps[1].join()

    elapsed = (time.monotonic() - start_time)
    assert elapsed >= .2

    ps[0].join()
    assert store.load('key') == -1


def test_fs_backend_releases_locks(tmpdir):
    # should deadlock if any method does not properly release its lock

    run_dir = tmpdir.mkdir('run_dir')
    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])

    def incr_from_other_process():
        other = Process(target=_fs_mp_increment_key, args=(run_dir, 'prefix', 'key', 0.))
        other.start()
        other.join()

    store.store('key', 42)
    incr_from_other_process()
    assert store.load('key') == 43

    store.load_store('key', lambda _: -1)
    incr_from_other_process()
    assert store.load('key') == 0

    incr_from_other_process()
    assert store.load('key') == 1


def _fs_mp_increment_key(run_dir, prefix, key, sleep):
    """Open a _FilesystemBackend and increment `key`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.  Sleeps for `sleep`
    seconds within the increment closure.
    """

    def l(x):
        time.sleep(sleep)
        return x + 1

    store = _FilesystemBackend(key_prefixes=[prefix], runtime_dirs=[run_dir])
    store.load_store(key, l)


def _fs_mp_check_key(run_dir, prefix, key, expected):
    """Open a _FilesystemBackend and check `key` value against `expected`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.
    """

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    assert store.load(key) == expected


def _fs_mp_store_key(run_dir, prefix, key, new_value):
    """Open a _FilesystemBackend and store `new_value` for `key`.

    For the `multiprocessing` tests.

    Opens the storage on `run_dir` and with `prefix`.
    """

    store = _FilesystemBackend(key_prefixes=['prefix'], runtime_dirs=[run_dir])
    store.store(key, new_value)



import pytest
from _testutils import MockPyusbDevice

from collections import deque

from liquidctl.driver.asetek_pro import HydroPro


@pytest.fixture
def emulate():
    usb_dev = MockPyusbDevice()
    cooler = HydroPro(usb_dev, 'Emulated Asetek Pro cooler', fan_count=2)
    return (usb_dev, cooler)


def test_not_totally_broken(emulate):
    _, cooler = emulate

    with cooler.connect():
        cooler.initialize()

        cooler.get_status()

        cooler.set_color(channel='logo', mode='blinking', colors=iter([[3, 2, 1]]))
        cooler.set_color(channel='logo', mode='pulse', colors=[[3, 2, 1]], speed='normal')

        cooler.set_speed_profile(channel='fan',
                                      profile=iter([(20, 20), (30, 50), (40, 100)]))
        cooler.set_speed_profile(channel='fan1',
                                      profile=iter([(20, 20), (30, 50), (40, 100)]))

        cooler.set_fixed_speed(channel='fan', duty=50)


def test_initialize_with_pump_mode(emulate):
    usb_dev, cooler = emulate

    with cooler.connect():
        cooler.initialize(pump_mode='balanced')

        _begin, set_pump, _fw_version = usb_dev._sent_xfers
        assert set_pump == ('write', 1, [0x32, 0x01])


def test_set_profile_of_all_fans(emulate):
    usb_dev, cooler = emulate

    # When setting the speed of all fans the driver first gets the speed of all
    # fans to work out how many fans are present. Set 2 valid responses to
    # simulate a setup with 2 fans present.  The first response is for the pump
    # setup in the initialize function.

    with cooler.connect():
        cooler.initialize()

        cooler.set_speed_profile(channel='fan', profile=[(0, 10), (25, 50), (40, 100)])

        _begin, _pump, _fw_version, fan1, fan2 = usb_dev._sent_xfers
        assert fan1 == ('write', 1, [0x40, 0x00, 0x00, 0x19, 0x28, 0x3c, 0x3c, 0x3c,
                                          0x3c, 0x0a, 0x32, 0x64, 0x64, 0x64, 0x64, 0x64])
        assert fan2 == ('write', 1, [0x40, 0x01, 0x00, 0x19, 0x28, 0x3c, 0x3c, 0x3c,
                                          0x3c, 0x0a, 0x32, 0x64, 0x64, 0x64, 0x64, 0x64])


def test_setting_speed_on_single_fan_2(emulate):
    usb_dev, cooler = emulate

    with cooler.connect():
        cooler.initialize()

        cooler.set_fixed_speed('fan2', 100)

        _begin, _pump, _fw_version, fan2 = usb_dev._sent_xfers
        assert fan2 == ('write', 1, [0x42, 0x01, 0x64])


def test_set_pump_to_fixed_color(emulate):
    usb_dev, cooler = emulate

    with cooler.connect():
        cooler.initialize()

        cooler.set_color(channel='logo', mode='fixed', colors=iter([[0xff, 0x88, 0x44]]))

        _begin, _pump, _fw_version, color_change, color_end = usb_dev._sent_xfers
        assert color_change == ('write', 1, [0x56, 0x02, 0xff, 0x88,
                                             0x44, 0xff, 0x88, 0x44])
        assert color_end == ('write', 1, [0x55, 0x01])


def test_set_pump_to_blinking_mode(emulate):
    usb_dev, cooler = emulate

    with cooler.connect():
        cooler.initialize()

        cooler.set_color(channel='logo', mode='blinking', speed='normal',
            colors=iter([[0xff, 0x88, 0x44], [0xff, 0xff, 0xff], [0x00, 0x00, 0x00]]))

        _begin, _pump, _fw_version, color_change, color_mode, color_end = usb_dev._sent_xfers
        assert color_change == ('write', 1, [0x56, 0x03, 0xff, 0x88, 0x44, 0xff,
                                             0xff, 0xff, 0x00, 0x00, 0x00])
        assert color_mode == ('write', 1, [0x53, 0x0A])
        assert color_end == ('write', 1, [0x58, 0x01])



"""LQiNFO – export monitoring data from liquidctl devices to HWiNFO.

This is a experimental script that exports the status of all available
devices to HWiNFO.

Usage:
  LQiNFO.py [options]
  LQiNFO.py --help
  LQiNFO.py --version

Options:
  --interval <seconds>    Update interval in seconds [default: 2]
  --legacy-690lc          Use Asetek 690LC in legacy mode (old Krakens)
  --vendor <id>           Filter devices by vendor id
  --product <id>          Filter devices by product id
  --release <number>      Filter devices by release number
  --serial <number>       Filter devices by serial number
  --bus <bus>             Filter devices by bus
  --address <address>     Filter devices by address in bus
  --usb-port <port>       Filter devices by USB port in bus
  --pick <number>         Pick among many results for a given filter
  -v, --verbose           Output additional information to stderr
  -g, --debug             Show debug information on stderr
  --version               Display the version number
  --help                  Show this message

Examples:
  python LQiNFO.py
  python LQiNFO.py --product 0xb200
  python LQiNFO.py --interval 0.5

Changelog:
  0.0.2  Fix cleanup of registry keys when exiting.
  0.0.1  First proof-of-concept.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import sys
import time
import winreg

import liquidctl.cli as _borrow
import usb

from docopt import docopt
from liquidctl.driver import *

LOGGER = logging.getLogger(__name__)

if __name__ == '__main__':
    args = docopt(__doc__, version='0.0.2')
    frwd = _borrow._make_opts(args)
    devs = list(find_liquidctl_devices(**frwd))
    update_interval = float(args['--interval'])

    if args['--debug']:
        args['--verbose'] = True
        logging.basicConfig(level=logging.DEBUG, format='[%(levelname)s] %(name)s: %(message)s')
    elif args['--verbose']:
        logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
    else:
        logging.basicConfig(level=logging.WARNING, format='%(message)s')
        sys.tracebacklimit = 0

    hwinfo_sensor_types = {
        '°C': 'Temp',
        'V': 'Volt',
        'rpm': 'Fan',
        'A': 'Current',
        'W': 'Power',
        'dB': 'Other'
    }
    hwinfo_custom = winreg.CreateKey(winreg.HKEY_CURRENT_USER, r'Software\HWiNFO64\Sensors\Custom')
    infos = []

    try:
        for i, d in enumerate(devs):
            LOGGER.info('Initializing %s', d.description)
            d.connect()

            dev_key = winreg.CreateKey(hwinfo_custom, f'{d.description} (liquidctl#{i})')
            dev_values = []
            for j, (k, v, u) in enumerate(d.get_status()):
                hwinfo_type = hwinfo_sensor_types.get(u, None)
                if not hwinfo_type:
                    dev_values.append(None)
                else:
                    sensor_key = winreg.CreateKey(dev_key, f'{hwinfo_type}{j}')
                    winreg.SetValueEx(sensor_key, 'Name', None, winreg.REG_SZ, k)
                    winreg.SetValueEx(sensor_key, 'Unit', None, winreg.REG_SZ, u)
                    winreg.SetValueEx(sensor_key, 'Value', None, winreg.REG_SZ, str(v))
                    dev_values.append(sensor_key)

            infos.append((dev_key, dev_values))
            # set up dev infos and registry
        status = {}
        while True:
            for i, d in enumerate(devs):
                dev_key, dev_values = infos[i]
                try:
                    for j, (k, v, u) in enumerate(d.get_status()):
                        sensor_key = dev_values[j]
                        if sensor_key:
                            winreg.SetValueEx(sensor_key, 'Value', None, winreg.REG_SZ, str(v))
                except usb.core.USBError as err:
                    LOGGER.warning('Failed to read from the device, possibly serving stale data')
                    LOGGER.debug(err, exc_info=True)
            time.sleep(update_interval)
    except KeyboardInterrupt:
        LOGGER.info('Canceled by user')
    except:
        LOGGER.exception('Unexpected error')
        sys.exit(1)
    finally:
        for d in devs:
            try:
                LOGGER.info('Disconnecting from %s', d.description)
                d.disconnect()
            except:
                LOGGER.exception('Unexpected error when disconnecting')
        for dev_key, dev_values in infos:
            try:
                for sensor_key in dev_values:
                    if sensor_key:
                        winreg.DeleteKey(sensor_key, '')
                winreg.DeleteKey(dev_key, '')
            except:
                LOGGER.exception('Unexpected error when cleaning the registry')



"""liquidctl drivers for Corsair Hydro Platinum and Pro XT liquid coolers.

Supported devices:

- Corsair Hydro H100i Platinum
- Corsair Hydro H100i Platinum SE
- Corsair Hydro H115i Platinum
- Corsair Hydro H60i Pro XT
- Corsair Hydro H100i Pro XT
- Corsair Hydro H115i Pro XT
- Corsair Hydro H150i Pro XT

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import itertools
import logging
import re
from enum import Enum, unique

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.keyval import RuntimeStorage
from liquidctl.util import RelaxedNamesEnum, clamp, fraction_of_byte, \
                           u16le_from, normalize_profile, mkCrcFun

_LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 64
_WRITE_PREFIX = 0x3f

_FEATURE_COOLING = 0b000
_FEATURE_COOLING2 = 0b011
_CMD_GET_STATUS = 0xff
_CMD_SET_COOLING = 0x14

_FEATURE_LIGHTING = None
_CMD_SET_LIGHTING1 = 0b100
_CMD_SET_LIGHTING2 = 0b101
_CMD_SET_LIGHTING3 = 0b110

# cooling data starts at offset 3 and ends just before the PEC byte
_SET_COOLING_DATA_LENGTH = _REPORT_LENGTH - 4
_SET_COOLING_DATA_PREFIX = [0x0, 0xff, 0x5, 0xff, 0xff, 0xff, 0xff, 0xff]
_FAN_MODE_OFFSETS = [0x0b - 3, 0x11 - 3]
_FAN_DUTY_OFFSETS = [offset + 5 for offset in _FAN_MODE_OFFSETS]
_FAN_PROFILE_OFFSETS = [0x1e - 3, 0x2c - 3]
_FAN_OFFSETS = list(zip(_FAN_MODE_OFFSETS, _FAN_DUTY_OFFSETS, _FAN_PROFILE_OFFSETS))
_PUMP_MODE_OFFSET = 0x17 - 3
_PROFILE_LENGTH_OFFSET = 0x1d - 3
_PROFILE_LENGTH = 7
_CRITICAL_TEMPERATURE = 60


@unique
class _FanMode(Enum):
    CUSTOM_PROFILE = 0x0
    CUSTOM_PROFILE_WITH_EXTERNAL_SENSOR = 0x1
    FIXED_DUTY = 0x2
    FIXED_RPM = 0x4

    @classmethod
    def _missing_(cls, value):
        _LOGGER.debug('falling back to FIXED_DUTY for _FanMode(%s)', value)
        return _FanMode.FIXED_DUTY


@unique
class _PumpMode(RelaxedNamesEnum):
    QUIET = 0x0
    BALANCED = 0x1
    EXTREME = 0x2

    @classmethod
    def _missing_(cls, value):
        _LOGGER.debug('falling back to BALANCED for _PumpMode(%s)', value)
        return _PumpMode.BALANCED


def _sequence(storage):
    """Return a generator that produces valid protocol sequence numbers.

    Sequence numbers increment across successful invocations of liquidctl, but
    are not atomic.  The sequence is: 1, 2, 3... 29, 30, 31, 1, 2, 3...

    In the protocol the sequence number is usually shifted left by 3 bits, and
    a shifted sequence will look like: 8, 16, 24... 232, 240, 248, 8, 16, 24...
    """

    while True:
        seq = storage.load_store('sequence', lambda x : x % 31 + 1, of_type=int, default=0)
        yield seq[1]


def _prepare_profile(original):
    clamped = ((temp, clamp(duty, 0, 100)) for temp, duty in original)
    normal = normalize_profile(clamped, _CRITICAL_TEMPERATURE)
    missing = _PROFILE_LENGTH - len(normal)
    if missing < 0:
        raise ValueError(f'too many points in profile (remove {-missing})')
    if missing > 0:
        normal += missing * [(_CRITICAL_TEMPERATURE, 100)]
    return normal


def _quoted(*names):
    return ', '.join(map(repr, names))


class HydroPlatinum(UsbHidDriver):
    """Corsair Hydro Platinum or Pro XT liquid cooler."""

    _MATCHES = [
        (0x1b1c, 0x0c18, 'Corsair Hydro H100i Platinum',
            {'fan_count': 2, 'fan_leds': 4}),
        (0x1b1c, 0x0c19, 'Corsair Hydro H100i Platinum SE',
            {'fan_count': 2, 'fan_leds': 16}),
        (0x1b1c, 0x0c17, 'Corsair Hydro H115i Platinum',
            {'fan_count': 2, 'fan_leds': 4}),
        (0x1b1c, 0x0c29, 'Corsair Hydro H60i Pro XT',
            {'fan_count': 2, 'fan_leds': 0}),
        (0x1b1c, 0x0c20, 'Corsair Hydro H100i Pro XT',
            {'fan_count': 2, 'fan_leds': 0}),
        (0x1b1c, 0x0c21, 'Corsair Hydro H115i Pro XT',
            {'fan_count': 2, 'fan_leds': 0}),
        (0x1b1c, 0x0c22, 'Corsair Hydro H150i Pro XT',
            {'fan_count': 3, 'fan_leds': 0}),
    ]

    @classmethod
    def probe(cls, handle, vendor=None, product=None, release=None,
              serial=None, match=None, **kwargs):
        """Probe `handle` and yield corresponding driver instances."""

        # this is modified from BaseUsbDriver.probe to match regardless of
        # presence of "Hydro", for backward compatibility with 1.5.0 and
        # previous versions

        for vid, pid, desc, devargs in cls._MATCHES:
            if (vendor and vendor != vid) or handle.vendor_id != vid:
                continue
            if (product and product != pid) or handle.product_id != pid:
                continue
            if release and handle.release_number != release:
                continue
            if serial and handle.serial_number != serial:
                continue
            if match:
                match = match.lower()
                descr = desc.lower()
                if not (match in descr or match in descr.replace('hydro ', '')):
                    continue
            consargs = devargs.copy()
            consargs.update(kwargs)
            dev = cls(handle, desc, **consargs)
            _LOGGER.debug('%s identified: %s', cls.__name__, desc)
            yield dev

    def __init__(self, device, description, fan_count, fan_leds, **kwargs):
        super().__init__(device, description, **kwargs)
        self._led_count = 16 + fan_count * fan_leds
        self._fan_names = [f'fan{i + 1}' for i in range(fan_count)]
        self._mincolors = {
            ('led', 'super-fixed'): 1,
            ('led', 'fixed'): 1,
            ('led', 'off'): 0,
        }
        self._maxcolors = {
            ('led', 'super-fixed'): self._led_count,
            ('led', 'fixed'): 1,
            ('led', 'off'): 0,
        }

        # the following fields are only initialized in connect()
        self._data = None
        self._sequence = None

    def connect(self, runtime_storage=None, **kwargs):
        """Connect to the device."""
        ret = super().connect(**kwargs)

        if runtime_storage:
            self._data = runtime_storage
        else:
            ids = f'vid{self.vendor_id:04x}_pid{self.product_id:04x}'
            # must use the HID path because there is no serial number; however,
            # these can be quite long on Windows and macOS, so only take the
            # numbers, since they are likely the only parts that vary between two
            # devices of the same model
            loc = 'loc' + '_'.join(re.findall(r'\d+', self.address))
            self._data = RuntimeStorage(key_prefixes=[ids, loc])

        self._sequence = _sequence(self._data)
        return ret

    def initialize(self, pump_mode='balanced', **kwargs):
        """Initialize the device and set the pump mode.

        The device should be initialized every time it is powered on, including when
        the system resumes from suspending to memory.

        Valid values for `pump_mode` are 'quiet', 'balanced' and 'extreme'.
        Unconfigured fan channels may default to 100% duty.  Subsequent calls
        should leave the fan speeds unaffected.

        Returns a list of `(property, value, unit)` tuples.
        """

        # set the flag so the LED command will need to be set again
        self._data.store('leds_enabled', 0)

        self._data.store('pump_mode', _PumpMode[pump_mode].value)
        res = self._send_set_cooling()
        fw_version = (res[2] >> 4, res[2] & 0xf, res[3])
        if fw_version < (1, 1, 0):
            # see: #201 ("Fan settings affects Fan 1 only and disables fan2")
            _LOGGER.warning('outdated and possibly unsupported firmware version')
        return [('Firmware version', '{}.{}.{}'.format(*fw_version), '')]

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        res = self._send_command(_FEATURE_COOLING, _CMD_GET_STATUS)

        info = [
            ('Liquid temperature', res[8] + res[7] / 255, '°C'),
        ]

        channels = [('Fan 1', 14), ('Fan 2', 21), ('Fan 3', 42)][:len(self._fan_names)]
        channels.append(('Pump', 28))

        for name, base in channels:
            info.append((f'{name} speed', u16le_from(res, offset=base + 1), 'rpm'))
            info.append((f'{name} duty', round(res[base] / 255 * 100), '%'))

        return info

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set fan or fans to a fixed speed duty.

        Valid channel values are 'fanN', where N >= 1 is the fan number, and
        'fan', to simultaneously configure all fans.  Unconfigured fan channels
        may default to 100% duty.
        """

        for hw_channel in self._get_hw_fan_channels(channel):
            self._data.store(f'{hw_channel}_mode', _FanMode.FIXED_DUTY.value)
            self._data.store(f'{hw_channel}_duty', duty)
        self._send_set_cooling()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Set fan or fans to follow a speed duty profile.

        Valid channel values are 'fanN', where N >= 1 is the fan number, and
        'fan', to simultaneously configure all fans.  Unconfigured fan channels
        may default to 100% duty.

        Up to seven (temperature, duty) pairs can be supplied in `profile`,
        with temperatures in Celsius and duty values in percentage.  The last
        point should set the fan to 100% duty cycle, or be omitted; in the
        latter case the fan will be set to max out at 60°C.
        """

        profile = list(profile)
        for hw_channel in self._get_hw_fan_channels(channel):
            self._data.store(f'{hw_channel}_mode', _FanMode.CUSTOM_PROFILE.value)
            self._data.store(f'{hw_channel}_profile', profile)
        self._send_set_cooling()

    def set_color(self, channel, mode, colors, **kwargs):
        """Set the color of each LED.

        In reality the device does not have the concept of different channels
        or modes, but this driver provides a few for convenience.  Animations
        still require successive calls to this API.

        The 'led' channel can be used to address individual LEDs, and supports
        the 'super-fixed', 'fixed' and 'off' modes.

        In 'super-fixed' mode, each color in `colors` is applied to one
        individual LED, successively.  LEDs for which no color has been
        specified default to off/solid black.  This is closest to how the
        device works.

        In 'fixed' mode, all LEDs are set to the first color taken from
        `colors`.  The `off` mode is equivalent to calling this function with
        'fixed' and a single solid black color in `colors`.

        The `colors` argument should be an iterable of one or more `[red, blue,
        green]` triples, where each red/blue/green component is a value in the
        range 0–255.

        The table bellow summarizes the available channels, modes, and their
        associated maximum number of colors for each device family.

        | Channel  | Mode        | LEDs         | Platinum | Pro XT | Platinum SE |
        | -------- | ----------- | ------------ | -------- | ------ | ----------- |
        | led      | off         | synchronized |        0 |      0 |           0 |
        | led      | fixed       | synchronized |        1 |      1 |           1 |
        | led      | super-fixed | independent  |       24 |     16 |          48 |
        """

        colors = list(colors)
        self._check_color_args(channel, mode, colors)
        if mode == 'off':
            expanded = []
        elif (channel, mode) == ('led', 'super-fixed'):
            expanded = colors[:self._led_count]
        elif (channel, mode) == ('led', 'fixed'):
            expanded = list(itertools.chain(*([color] * self._led_count for color in colors[:1])))
        else:
            assert False, 'assumed unreacheable'

        if self._data.load('leds_enabled', of_type=int, default=0) == 0:
            # These hex strings are currently magic values that work but Im not quite sure why.
            d1 = bytes.fromhex("0101ffffffffffffffffffffffffff7f7f7f7fff00ffffffff00ffffffff00ffffffff00ffffffff00ffffffff00ffffffffffffffffffffffffffffff")
            d2 = bytes.fromhex("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f2021222324252627ffffffffffffffffffffffffffffffffffffffffff")
            d3 = bytes.fromhex("28292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f404142434445464748494a4b4c4d4e4fffffffffffffffffffffffffffffffffffffffffff")

            # Send the magic messages to enable setting the LEDs to statuC values
            self._send_command(None, 0b001, data=d1)
            self._send_command(None, 0b010, data=d2)
            self._send_command(None, 0b011, data=d3)
            self._data.store('leds_enabled', 1)

        data1 = bytes(itertools.chain(*((b, g, r) for r, g, b in expanded[0:20])))
        data2 = bytes(itertools.chain(*((b, g, r) for r, g, b in expanded[20:40])))
        data3 = bytes(itertools.chain(*((b, g, r) for r, g, b in expanded[40:])))

        self._send_command(_FEATURE_LIGHTING, _CMD_SET_LIGHTING1, data=data1)

        if self._led_count > 20:
            self._send_command(_FEATURE_LIGHTING, _CMD_SET_LIGHTING2, data=data2)

        if self._led_count > 40:
            self._send_command(_FEATURE_LIGHTING, _CMD_SET_LIGHTING3, data=data3)

    def _check_color_args(self, channel, mode, colors):
        try:
            mincolors = self._mincolors[(channel, mode)]
            maxcolors = self._maxcolors[(channel, mode)]
        except KeyError:
            raise ValueError(f'unsupported (channel, mode) pair, '
                             f'should be one of: {_quoted(*self._mincolors)}') from None
        if len(colors) < mincolors:
            raise ValueError(f'at least {mincolors} required for {_quoted(channel, mode)}')
        if len(colors) > maxcolors:
            _LOGGER.warning('too many colors, dropping to %d', maxcolors)
            return maxcolors
        return len(colors)

    def _get_hw_fan_channels(self, channel):
        if channel == 'fan':
            return self._fan_names
        if channel in self._fan_names:
            return [channel]
        raise ValueError(f'unknown channel, should be one of: {_quoted("fan", *self._fan_names)}')

    def _send_command(self, feature, command, data=None):
        # self.device.write expects buf[0] to be the report number or 0 if not used
        buf = bytearray(_REPORT_LENGTH + 1)
        buf[1] = _WRITE_PREFIX
        buf[2] = next(self._sequence) << 3
        if feature is not None:
            buf[2] |= feature
            buf[3] = command
            start_at = 4
        else:
            buf[2] |= command
            start_at = 3
        if data:
            buf[start_at: start_at + len(data)] = data
        buf[-1] = mkCrcFun('crc-8')(buf[2:-1])
        self.device.clear_enqueued_reports()
        self.device.write(buf)
        buf = bytes(self.device.read(_REPORT_LENGTH))
        if mkCrcFun('crc-8')(buf[1:]):
            _LOGGER.warning('response checksum does not match data')
        return buf

    def _generate_cooling_payload(self, fan_names):

        data = bytearray(_SET_COOLING_DATA_LENGTH)
        data[0: len(_SET_COOLING_DATA_PREFIX)] = _SET_COOLING_DATA_PREFIX
        data[_PROFILE_LENGTH_OFFSET] = _PROFILE_LENGTH

        for fan, (imode, iduty, iprofile) in zip(fan_names, _FAN_OFFSETS):
            mode = _FanMode(self._data.load(f'{fan}_mode', of_type=int))
            if mode is _FanMode.FIXED_DUTY:
                stored = self._data.load(f'{fan}_duty', of_type=int, default=100)
                duty = clamp(stored, 0, 100)
                data[iduty] = fraction_of_byte(percentage=duty)
                _LOGGER.info('setting %s to %d%% duty cycle', fan, duty)
            elif mode is _FanMode.CUSTOM_PROFILE:
                stored = self._data.load(f'{fan}_profile', of_type=list, default=[])
                profile = _prepare_profile(stored)  # ensures correct len(profile)
                pairs = ((temp, fraction_of_byte(percentage=duty)) for temp, duty in profile)
                data[iprofile: iprofile + _PROFILE_LENGTH * 2] = itertools.chain(*pairs)
                _LOGGER.info('setting %s to follow profile %r', fan, profile)
            else:
                raise ValueError(f'unsupported fan {mode}')
            data[imode] = mode.value

        return data

    def _send_set_cooling(self):
        pump_mode = _PumpMode(self._data.load('pump_mode', of_type=int))

        data = self._generate_cooling_payload(self._fan_names[0:2])
        data[_PUMP_MODE_OFFSET] = pump_mode.value
        _LOGGER.info('setting pump mode to %s', pump_mode.name.lower())

        if len(self._fan_names) == 3:
            data2 = self._generate_cooling_payload(self._fan_names[2:])
            data2[_PUMP_MODE_OFFSET] = 0xff
            self._send_command(_FEATURE_COOLING2, _CMD_SET_COOLING, data=data2)

        return self._send_command(_FEATURE_COOLING, _CMD_SET_COOLING, data=data)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()



import sys

collect_ignore = ['setup.py']
collect_ignore.append('extra/contrib/fusion_rgb_cycle.py')  # depends on coloraide

if sys.platform != 'linux':
    collect_ignore.append('tests/test_smbus.py')

if sys.platform not in ['win32', 'cygwin']:
    collect_ignore.append('extra/windows/LQiNFO.py')



# uses the psf/black style

import sys
from inspect import cleandoc

if __name__ == "__main__":

    # This script is meant to be executed from this directory or the project root.
    # We use that assumption to make Python pick the local liquidctl modules,
    # instead other versions that may be installed on the environment/system.
    sys.path = ["../..", ""] + sys.path

    # The 71-liquidctl.rules file must always be written in UTF-8.  Therefore,
    # when we detect that stdout has been redirected to a file, make sure that
    # it is set to UTF-8, regardless of the current locale.
    if not sys.stdout.isatty() and sys.stdout.encoding != "utf-8":
        print("liquidctl udev rules files must use UTF-8, reconfiguring stdout", file=sys.stderr)
        sys.stdout.reconfigure(encoding="utf-8")

    from liquidctl.driver.base import find_all_subclasses
    from liquidctl.driver.nvidia import _NvidiaI2CDriver
    from liquidctl.driver.usb import BaseUsbDriver

    HEADER = """
    # Rules that grant unprivileged access to devices supported by liquidctl.
    #
    # Users and distros are encouraged to use these if they want liquidctl to work
    # without requiring root privileges (e.g. without the use of `sudo`).
    #
    # In the case of I²C/SMBus devices, these rules also cause the loading of the
    # `i2c-dev` kernel module.  This module is required for access to I²C/SMBus
    # devices from userspace, and manually loading kernel modules is in itself a
    # privileged operation.
    #
    # Distros will likely want to place this file in `/usr/lib/udev/rules.d/`, while
    # users installing this manually SHOULD use `/etc/udev/rules.d/` instead.
    #
    # The suggested name for this file is `71-liquidctl.rules`.  Whatever name is
    # used, it MUST lexically appear before 73-seat-late.rules.  The suggested name
    # was chosen so that it is also lexically after systemd-provided 70-uaccess.rules.
    #
    # Once installed, reload and trigger the new rules with:
    #
    #   # udevadm control --reload
    #   # udevadm trigger
    #
    # Note that this will not change the mode of `/dev/hidraw*` devices that have
    # already been created.  In practice, this means that HIDs may continue to require
    # privileged access until, either, they are rebound to their kernel drivers, or
    # the system is rebooted.
    #
    # These rules assume a system with modern versions of systemd/udev, that support
    # the `uaccess` tag.  On older systems the rules can be changed to instead set
    # GROUP="plugdev" and MODE="0660"; other groups and modes may also be used.
    #
    # The use of the `uaccess` mechanism assumes that only physical sessions (or
    # "seats") need unprivileged access to the devices.[^1][^2]  In case headless
    # sessions are also expected to interactively run liquidctl, GROUP and MODE should
    # also be set, as a fallback.
    #
    # Finally, this file was automatically generated.  To update it, from a Linux
    # shell and the current directory, execute:
    #
    #     $ python generate-uaccess-udev-rules.py > 71-liquidctl.rules
    #
    # [^1]: https://github.com/systemd/systemd/issues/4288
    # [^2]: https://wiki.archlinux.org/title/Users_and_groups#Pre-systemd_groups
    """

    MANUAL_RULES = r"""
        # Section: special cases

        # Host SMBus on Intel mainstream/HEDT platforms
        KERNEL=="i2c-*", DRIVERS=="i801_smbus", TAG+="uaccess", \
            RUN{builtin}="kmod load i2c-dev"
    """

    print(cleandoc(HEADER))

    print()
    print()
    print(cleandoc(MANUAL_RULES))

    print()
    print()
    print(f"# Section: NVIDIA graphics cards")

    nvidia_devs = {}

    for driver in find_all_subclasses(_NvidiaI2CDriver):
        for did, sdid, description in driver._MATCHES:
            ids = (driver._VENDOR, did, sdid)
            if ids in nvidia_devs:
                nvidia_devs[ids].append(description)
                nvidia_devs[ids].sort()
            else:
                nvidia_devs[ids] = [description]

    nvidia_devs = [
        (svid, did, sdid, description) for (svid, did, sdid), description in nvidia_devs.items()
    ]
    nvidia_devs.sort(key=lambda x: x[3][0])

    for svid, did, sdid, descriptions in nvidia_devs:
        print()
        for desc in descriptions:
            desc = desc.replace(" (experimental)", "")
            print(f"# {desc}")
        entry = f"""
            KERNEL=="i2c-*", ATTR{{name}}=="NVIDIA i2c adapter 1 *", ATTRS{{vendor}}=="0x10de", \\
                ATTRS{{device}}=="{did:#06x}", ATTRS{{subsystem_vendor}}=="{svid:#06x}", \\
                ATTRS{{subsystem_device}}=="{sdid:#06x}", DRIVERS=="nvidia", TAG+="uaccess", \\
                RUN{{builtin}}="kmod load i2c-dev"
        """
        print(cleandoc(entry))

    print()
    print()
    print(f"# Section: USB devices and USB HIDs")

    usb_devs = {}

    for driver in find_all_subclasses(BaseUsbDriver):
        for vid, pid, description, _ in driver._MATCHES:
            ids = (vid, pid)
            if ids in usb_devs:
                usb_devs[ids].append(description)
                usb_devs[ids].sort()
            else:
                usb_devs[ids] = [description]

    usb_devs = [(vid, pid, description) for (vid, pid), description in usb_devs.items()]
    usb_devs.sort(key=lambda x: x[2][0])

    for vid, pid, descriptions in usb_devs:
        print()
        for desc in descriptions:
            desc = desc.replace(" (experimental)", "")
            print(f"# {desc}")
        print(
            f'SUBSYSTEMS=="usb", ATTRS{{idVendor}}=="{vid:04x}", ATTRS{{idProduct}}=="{pid:04x}", TAG+="uaccess"'
        )



"""Interfaces to read the liquidctl version.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style


# keep setuptools_scm parameters in sync with pyproject.toml
_SETUPTOOLS_SCM_PARAMS = {"version_scheme": "release-branch-semver"}


def _build_version():

    try:
        from liquidctl._version import version, version_tuple

        return (version, version_tuple)
    except ModuleNotFoundError:
        return None


def _runtime_version():

    try:
        from setuptools_scm import get_version
    except ModuleNotFoundError:
        return None

    # first, assume that we're a git checkout
    try:
        version = get_version(**_SETUPTOOLS_SCM_PARAMS)
        if version:
            return (version, None)
    except LookupError:
        pass

    # if that also failed, assume that we're a tarball
    try:
        guess = get_version(parentdir_prefix_version="liquidctl-", **_SETUPTOOLS_SCM_PARAMS)
        if guess:
            if "+" in guess:
                guess += "-guessed"
            else:
                guess += "+guessed"
            return (guess, None)
    except LookupError:
        pass

    return None


# - try to get the version written by setuptools_scm during the build;
# - otherwise try to compute one right now;
# - failing that too, use an obviously invalid value
#
# (_version_tuple is kept private as it's only available in some cases and
# don't want to commit to it yet)
(version, _version_tuple) = _build_version() or _runtime_version() or ("0.0.0-unknown", None)

# old field name (liquidctl.__version__ is preferred now)
__version__ = version
